# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the termination round classes."""

from enum import Enum
from typing import Dict, Optional, Set, Tuple, cast

from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)
from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    BaseTxPayload,
    CollectSameUntilThresholdRound,
    TransactionNotValidError,
    get_name,
)
from packages.valory.skills.termination_abci.payloads import BackgroundPayload
from packages.valory.skills.transaction_settlement_abci.rounds import (
    FailedRound,
    FinishedTransactionSubmissionRound,
    TransactionSubmissionAbciApp,
)


class Event(Enum):
    """Defines the (background) round events."""

    TERMINATE = "terminate"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def termination_majority_reached(self) -> bool:
        """Get termination_majority_reached."""
        return cast(bool, self.db.get("termination_majority_reached", False))

    @property
    def most_voted_tx_hash(self) -> str:
        """Get the most_voted_tx_hash."""
        return cast(str, self.db.get_strict("most_voted_tx_hash"))  # pragma: no cover


class BackgroundRound(CollectSameUntilThresholdRound):
    """Defines the background round, which runs concurrently with other rounds."""

    payload_class = BackgroundPayload
    synchronized_data_class = SynchronizedData

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""
        # for background round payloads, we don't need to check the round_count, as round_count is irrelevant for the
        # background since it's running concurrently in the background.
        sender = payload.sender
        if sender not in self.accepting_payloads_from:
            raise ABCIAppInternalError(
                f"{sender} not in list of participants: {sorted(self.accepting_payloads_from)}"
            )

        if sender in self.collection:
            raise ABCIAppInternalError(
                f"sender {sender} has already sent value for round: {self.round_id}"
            )

        self.collection[sender] = payload

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check Payload"""
        # NOTE: the TransactionNotValidError is intercepted in ABCIRoundHandler.deliver_tx
        #  which means it will be logged instead of raised
        # for background round payloads, we don't need to check the round_count, as round_count is irrelevant for the
        # background since it's running concurrently in the background.
        sender_in_participant_set = payload.sender in self.accepting_payloads_from
        if not sender_in_participant_set:
            raise TransactionNotValidError(
                f"{payload.sender} not in list of participants: {sorted(self.accepting_payloads_from)}"
            )

        if payload.sender in self.collection:
            raise TransactionNotValidError(
                f"sender {payload.sender} has already sent value for round: {self.round_id}"
            )

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.threshold_reached:
            state = self.synchronized_data.update(
                synchronized_data_class=self.synchronized_data_class,
                **{
                    get_name(SynchronizedData.termination_majority_reached): True,
                    get_name(
                        SynchronizedData.most_voted_tx_hash
                    ): self.most_voted_payload,
                },
            )
            return state, Event.TERMINATE

        return None


class TerminationRound(AbstractRound):
    """Round to act as the counterpart of the behaviour responsible for terminating the agent."""

    payload_class = None
    synchronized_data_class = SynchronizedData

    def check_payload(self, payload: BaseTxPayload) -> None:
        """No logic required here."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """No logic required here."""

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """No logic required here."""
        return None


class PostTerminationTxAbciApp(AbciApp[Event]):
    """PostTerminationTxAbciApp

    Initial round: TerminationRound

    Initial states: {TerminationRound}

    Transition states:
        0. TerminationRound
            - terminate: 0.

    Final states: {}

    Timeouts:

    """

    initial_round_cls = TerminationRound
    # the following is not needed, it is added to satisfy the round check
    # the TerminationRound when run it terminates the agent, so nothing can come after it
    transition_function = {TerminationRound: {Event.TERMINATE: TerminationRound}}
    db_pre_conditions: Dict[AppState, Set[str]] = {TerminationRound: set()}


termination_transition_function: AbciAppTransitionMapping = {
    FinishedTransactionSubmissionRound: PostTerminationTxAbciApp.initial_round_cls,
    FailedRound: TransactionSubmissionAbciApp.initial_round_cls,
}
TerminationAbciApp = chain(
    (
        TransactionSubmissionAbciApp,
        PostTerminationTxAbciApp,
    ),
    termination_transition_function,
)

TerminationAbciApp.transition_function[BackgroundRound] = {
    Event.TERMINATE: TransactionSubmissionAbciApp.initial_round_cls,
}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/registration_abci skill's behaviours."""

# pylint: skip-file

import logging
import time
from collections import deque
from pathlib import Path
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from unittest import mock
from unittest.mock import MagicMock

import pytest
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedMessage,
    SignedTransaction,
)
from aea.helpers.transaction.base import State as TrState
from aea.helpers.transaction.base import TransactionDigest, TransactionReceipt
from aea.skills.base import SkillContext
from web3.types import Nonce

from packages.open_aea.protocols.signing import SigningMessage
from packages.valory.contracts.gnosis_safe.contract import (
    PUBLIC_ID as GNOSIS_SAFE_CONTRACT_ID,
)
from packages.valory.protocols.abci import AbciMessage  # noqa: F401
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    RPCResponseStatus,
    make_degenerate_behaviour,
)
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.abstract_round_abci.test_tools.common import (
    BaseRandomnessBehaviourTest,
    BaseSelectKeeperBehaviourTest,
)
from packages.valory.skills.transaction_settlement_abci import PUBLIC_ID
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    CheckLateTxHashesBehaviour,
    CheckTransactionHistoryBehaviour,
    FinalizeBehaviour,
    REVERT_CODES_TO_REASONS,
    RandomnessTransactionSubmissionBehaviour,
    ResetBehaviour,
    SelectKeeperTransactionSubmissionBehaviourA,
    SelectKeeperTransactionSubmissionBehaviourB,
    SignatureBehaviour,
    SynchronizeLateMessagesBehaviour,
    TransactionSettlementBaseBehaviour,
    TxDataType,
    ValidateTransactionBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    hash_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    Event as TransactionSettlementEvent,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    FinishedTransactionSubmissionRound,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TransactionSettlementSynchronizedSata,
)


PACKAGE_DIR = Path(__file__).parent.parent


def mock_yield_and_return(
    return_value: Any,
) -> Callable[[], Generator[None, None, Any]]:
    """Wrapper for a Dummy generator that returns a `bool`."""

    def yield_and_return(*_: Any, **__: Any) -> Generator[None, None, Any]:
        """Dummy generator that returns a `bool`."""
        yield
        return return_value

    return yield_and_return


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


class TransactionSettlementFSMBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing TransactionSettlement FSMBehaviour."""

    path_to_skill = PACKAGE_DIR

    def ffw_signature(self, db_items: Optional[Dict] = None) -> None:
        """Fast-forward to the `SignatureBehaviour`."""
        if db_items is None:
            db_items = {}

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=SignatureBehaviour.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(db_items),
                )
            ),
        )


class TestTransactionSettlementBaseBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test `TransactionSettlementBaseBehaviour`."""

    @pytest.mark.parametrize(
        "message, tx_digest, rpc_status, expected_data, replacement",
        (
            (
                MagicMock(
                    performative=ContractApiMessage.Performative.ERROR, message="GS026"
                ),
                None,
                RPCResponseStatus.SUCCESS,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.VERIFIED,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(
                    performative=ContractApiMessage.Performative.ERROR, message="test"
                ),
                None,
                RPCResponseStatus.SUCCESS,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.ERROR,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_MESSAGE),
                None,
                RPCResponseStatus.SUCCESS,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_TRANSACTION),
                None,
                RPCResponseStatus.INCORRECT_NONCE,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.ERROR,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_TRANSACTION),
                None,
                RPCResponseStatus.INSUFFICIENT_FUNDS,
                {
                    "blacklisted_keepers": {"agent_1" + "-" * 35},
                    "keeper_retries": 1,
                    "keepers": deque(("agent_3" + "-" * 35,)),
                    "status": VerificationStatus.INSUFFICIENT_FUNDS,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_TRANSACTION),
                None,
                RPCResponseStatus.UNCLASSIFIED_ERROR,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_TRANSACTION),
                None,
                RPCResponseStatus.UNDERPRICED,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "",
                },
                False,
            ),
            (
                MagicMock(performative=ContractApiMessage.Performative.RAW_TRANSACTION),
                "test_digest_0",
                RPCResponseStatus.ALREADY_KNOWN,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "test_digest_0",
                },
                False,
            ),
            (
                MagicMock(
                    performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                    raw_transaction=MagicMock(
                        body={
                            "nonce": 0,
                            "maxPriorityFeePerGas": 10,
                            "maxFeePerGas": 20,
                            "gas": 0,
                        }
                    ),
                ),
                "test_digest_1",
                RPCResponseStatus.SUCCESS,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "test_digest_1",
                },
                False,
            ),
            (
                MagicMock(
                    performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                    raw_transaction=MagicMock(
                        body={
                            "nonce": 0,
                            "maxPriorityFeePerGas": 10,
                            "maxFeePerGas": 20,
                            "gas": 0,
                        }
                    ),
                ),
                "test_digest_2",
                RPCResponseStatus.SUCCESS,
                {
                    "blacklisted_keepers": set(),
                    "keeper_retries": 2,
                    "keepers": deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)),
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "test_digest_2",
                },
                True,
            ),
        ),
    )
    def test__get_tx_data(
        self,
        message: ContractApiMessage,
        tx_digest: Optional[str],
        rpc_status: RPCResponseStatus,
        expected_data: TxDataType,
        replacement: bool,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test `_get_tx_data`."""
        # fast-forward to any behaviour of the tx settlement skill
        init_db_items = dict(
            most_voted_tx_hash="b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d90000000"
            "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
            "0000000000000000000000002625a000x77E9b2EF921253A171Fa0CB9ba80558648Ff7215b0e6add595e00477c"
            "f347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283"
            "ac76e4efce2a674fe72d9",
            keepers=int(2).to_bytes(32, "big").hex()
            + "".join(deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35))),
        )
        self.ffw_signature(init_db_items)
        behaviour = cast(SignatureBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == SignatureBehaviour.auto_behaviour_id()
        # Set `nonce` to the same value as the returned, so that we test the tx replacement logging.
        if replacement:
            behaviour.params.mutable_params.nonce = Nonce(0)

        # patch the `send_raw_transaction` method
        def dummy_send_raw_transaction(
            *_: Any,
        ) -> Generator[None, None, Tuple[Optional[str], RPCResponseStatus]]:
            """Dummy `send_raw_transaction` method."""
            yield
            return tx_digest, rpc_status

        monkeypatch.setattr(
            BaseBehaviour, "send_raw_transaction", dummy_send_raw_transaction
        )
        # call `_get_tx_data`
        tx_data_iterator = cast(
            TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
        )._get_tx_data(message, use_flashbots=False)

        if message.performative == ContractApiMessage.Performative.RAW_TRANSACTION:
            next(tx_data_iterator)

        try:
            next(tx_data_iterator)
        except StopIteration as res:
            assert res.value == expected_data

        """Test the serialized_keepers method."""
        behaviour_ = self.behaviour.current_behaviour
        assert behaviour_ is not None
        assert behaviour_.serialized_keepers(deque([]), 1) == ""
        assert (
            behaviour_.serialized_keepers(deque(["-" * 42]), 1)
            == "0000000000000000000000000000000000000000000000000000000000000001"
            "------------------------------------------"
        )

    @pytest.mark.parametrize(
        argnames=["tx_body", "expected_params"],
        argvalues=[
            [
                {"maxPriorityFeePerGas": "dummy", "maxFeePerGas": "dummy"},
                ["maxPriorityFeePerGas", "maxFeePerGas"],
            ],
            [{"gasPrice": "dummy"}, ["gasPrice"]],
            [
                {"maxPriorityFeePerGas": "dummy"},
                [],
            ],
            [
                {"maxFeePerGas": "dummy"},
                [],
            ],
            [
                {},
                [],
            ],
            [
                {
                    "maxPriorityFeePerGas": "dummy",
                    "maxFeePerGas": "dummy",
                    "gasPrice": "dummy",
                },
                ["maxPriorityFeePerGas", "maxFeePerGas"],
            ],
        ],
    )
    def test_get_gas_price_params(
        self, tx_body: dict, expected_params: List[str]
    ) -> None:
        """Test the get_gas_price_params method"""
        # fast-forward to any behaviour of the tx settlement skill
        self.ffw_signature()

        assert (
            cast(
                TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
            ).get_gas_price_params(tx_body)
            == expected_params
        )

    def test_parse_revert_reason_successful(self) -> None:
        """Test `_parse_revert_reason` method."""
        # fast-forward to any behaviour of the tx settlement skill
        self.ffw_signature()

        for code, explanation in REVERT_CODES_TO_REASONS.items():
            message = MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message=f"some text {code}.",
            )

            expected = f"Received a {code} revert error: {explanation}."

            assert (
                cast(
                    TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
                )._parse_revert_reason(message)
                == expected
            )

    @pytest.mark.parametrize(
        "message",
        (
            MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message="Non existing code should be invalid GS086.",
            ),
            MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message="Code not matching the regex should be invalid GS0265.",
            ),
            MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message="No code in the message should be invalid.",
            ),
            MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message="",  # empty message should be invalid
            ),
            MagicMock(
                performative=ContractApiMessage.Performative.ERROR,
                message=None,  # `None` message should be invalid
            ),
        ),
    )
    def test_parse_revert_reason_unsuccessful(
        self, message: ContractApiMessage
    ) -> None:
        """Test `_parse_revert_reason` method."""
        # fast-forward to any behaviour of the tx settlement skill
        self.ffw_signature()

        expected = f"get_raw_safe_transaction unsuccessful! Received: {message}"

        assert (
            cast(
                TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
            )._parse_revert_reason(message)
            == expected
        )


class TestRandomnessInOperation(BaseRandomnessBehaviourTest):
    """Test randomness in operation."""

    path_to_skill = PACKAGE_DIR

    randomness_behaviour_class = RandomnessTransactionSubmissionBehaviour
    next_behaviour_class = SelectKeeperTransactionSubmissionBehaviourA
    done_event = TransactionSettlementEvent.DONE


class TestSelectKeeperTransactionSubmissionBehaviourA(BaseSelectKeeperBehaviourTest):
    """Test SelectKeeperBehaviour."""

    path_to_skill = PACKAGE_DIR

    select_keeper_behaviour_class = SelectKeeperTransactionSubmissionBehaviourA
    next_behaviour_class = SignatureBehaviour
    done_event = TransactionSettlementEvent.DONE
    _synchronized_data = TransactionSettlementSynchronizedSata


class TestSelectKeeperTransactionSubmissionBehaviourB(
    TestSelectKeeperTransactionSubmissionBehaviourA
):
    """Test SelectKeeperBehaviour."""

    select_keeper_behaviour_class = SelectKeeperTransactionSubmissionBehaviourB
    next_behaviour_class = FinalizeBehaviour

    @mock.patch.object(
        TransactionSettlementSynchronizedSata,
        "keepers",
        new_callable=mock.PropertyMock,
    )
    @mock.patch.object(
        TransactionSettlementSynchronizedSata,
        "keeper_retries",
        new_callable=mock.PropertyMock,
    )
    @mock.patch.object(
        TransactionSettlementSynchronizedSata,
        "final_verification_status",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        "keepers, keeper_retries, blacklisted_keepers, final_verification_status",
        (
            (
                deque(f"keeper_{i}" for i in range(4)),
                1,
                set(),
                VerificationStatus.NOT_VERIFIED,
            ),
            (deque(("test_keeper",)), 2, set(), VerificationStatus.PENDING),
            (deque(("test_keeper",)), 2, set(), VerificationStatus.NOT_VERIFIED),
            (deque(("test_keeper",)), 2, {"a1"}, VerificationStatus.NOT_VERIFIED),
            (
                deque(("test_keeper",)),
                2,
                {"test_keeper"},
                VerificationStatus.NOT_VERIFIED,
            ),
            (
                deque(("test_keeper",)),
                2,
                {"a_1", "a_2", "test_keeper"},
                VerificationStatus.NOT_VERIFIED,
            ),
            (deque(("test_keeper",)), 1, set(), VerificationStatus.NOT_VERIFIED),
            (deque(("test_keeper",)), 3, set(), VerificationStatus.NOT_VERIFIED),
        ),
    )
    def test_select_keeper(
        self,
        final_verification_status_mock: mock.PropertyMock,
        keeper_retries_mock: mock.PropertyMock,
        keepers_mock: mock.PropertyMock,
        keepers: Deque[str],
        keeper_retries: int,
        blacklisted_keepers: Set[str],
        final_verification_status: VerificationStatus,
    ) -> None:
        """Test select keeper agent."""
        keepers_mock.return_value = keepers
        keeper_retries_mock.return_value = keeper_retries
        final_verification_status_mock.return_value = final_verification_status
        super().test_select_keeper(blacklisted_keepers=blacklisted_keepers)

    @mock.patch.object(
        TransactionSettlementSynchronizedSata,
        "final_verification_status",
        new_callable=mock.PropertyMock,
        return_value=VerificationStatus.PENDING,
    )
    @pytest.mark.skip  # Needs to be investigated, fails in CI only. look at #1710
    def test_select_keeper_tx_pending(
        self, _: mock.PropertyMock, caplog: LogCaptureFixture
    ) -> None:
        """Test select keeper while tx is pending"""

        with caplog.at_level(logging.INFO):
            super().test_select_keeper(blacklisted_keepers=set())
            assert "Kept keepers and incremented retries" in caplog.text


class TestSignatureBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test SignatureBehaviour."""

    def test_signature_behaviour(
        self,
    ) -> None:
        """Test signature behaviour."""

        init_db_items = dict(
            most_voted_tx_hash="b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d90000000"
            "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
            "0000000000000000000000002625a000x77E9b2EF921253A171Fa0CB9ba80558648Ff7215b0e6add595e00477c"
            "f347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283"
            "ac76e4efce2a674fe72d9",
        )
        self.ffw_signature(init_db_items)

        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == SignatureBehaviour.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_signing_request(
            request_kwargs=dict(
                performative=SigningMessage.Performative.SIGN_MESSAGE,
            ),
            response_kwargs=dict(
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SignedMessage(
                    ledger_id="ethereum", body="stub_signature"
                ),
            ),
        )
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == FinalizeBehaviour.auto_behaviour_id()


class TestFinalizeBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test FinalizeBehaviour."""

    behaviour_class = FinalizeBehaviour

    def test_non_sender_act(
        self,
    ) -> None:
        """Test finalize behaviour."""
        participants = (self.skill.skill_context.agent_address, "a_1", "a_2")
        retries = 1
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            most_voted_keeper_address="most_voted_keeper_address",
                            participants=participants,
                            # keeper needs to have length == 42 in order to be parsed
                            keepers=retries.to_bytes(32, "big").hex()
                            + "other_agent"
                            + "-" * 31,
                        )
                    ),
                )
            ),
        )
        assert self.behaviour.current_behaviour is not None
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        cast(
            FinalizeBehaviour, self.behaviour.current_behaviour
        ).params.mutable_params.tx_hash = "test"
        self.behaviour.act_wrapper()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(ValidateTransactionBehaviour, self.behaviour.current_behaviour)
        assert (
            behaviour.behaviour_id == ValidateTransactionBehaviour.auto_behaviour_id()
        )
        assert behaviour.params.mutable_params.tx_hash == "test"

    @pytest.mark.parametrize(
        "resubmitting, response_kwargs",
        (
            (
                (
                    True,
                    dict(
                        performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                        callable="get_deploy_transaction",
                        raw_transaction=RawTransaction(
                            ledger_id="ethereum",
                            body={
                                "tx_hash": "0x3b",
                                "nonce": 0,
                                "maxFeePerGas": int(10e10),
                                "maxPriorityFeePerGas": int(10e10),
                            },
                        ),
                    ),
                )
            ),
            (
                False,
                dict(
                    performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                    callable="get_deploy_transaction",
                    raw_transaction=RawTransaction(
                        ledger_id="ethereum",
                        body={
                            "tx_hash": "0x3b",
                            "nonce": 0,
                            "maxFeePerGas": int(10e10),
                            "maxPriorityFeePerGas": int(10e10),
                        },
                    ),
                ),
            ),
            (
                False,
                dict(
                    performative=ContractApiMessage.Performative.ERROR,
                    callable="get_deploy_transaction",
                    code=500,
                    message="GS026",
                    data=b"",
                ),
            ),
            (
                False,
                dict(
                    performative=ContractApiMessage.Performative.ERROR,
                    callable="get_deploy_transaction",
                    code=500,
                    message="other error",
                    data=b"",
                ),
            ),
        ),
    )
    @mock.patch.object(SkillContext, "agent_address", new_callable=mock.PropertyMock)
    def test_sender_act(
        self,
        agent_address_mock: mock.PropertyMock,
        resubmitting: bool,
        response_kwargs: Dict[
            str,
            Union[
                int,
                str,
                bytes,
                Dict[str, Union[int, str]],
                ContractApiMessage.Performative,
                RawTransaction,
            ],
        ],
    ) -> None:
        """Test finalize behaviour."""
        nonce: Optional[int] = None
        max_priority_fee_per_gas: Optional[int] = None

        if resubmitting:
            nonce = 0
            max_priority_fee_per_gas = 1

        # keepers need to have length == 42 in order to be parsed
        agent_address_mock.return_value = "-" * 42
        retries = 1
        participants = (
            self.skill.skill_context.agent_address,
            "a_1" + "-" * 39,
            "a_2" + "-" * 39,
        )
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            safe_contract_address="safe_contract_address",
                            participants=participants,
                            participant_to_signature={},
                            most_voted_tx_hash=hash_payload_to_hex(
                                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                                1,
                                1,
                                "0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215",
                                b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                            ),
                            nonce=nonce,
                            max_priority_fee_per_gas=max_priority_fee_per_gas,
                            keepers=retries.to_bytes(32, "big").hex()
                            + self.skill.skill_context.agent_address,
                        )
                    ),
                )
            ),
        )

        assert self.behaviour.current_behaviour is not None
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        cast(
            FinalizeBehaviour, self.behaviour.current_behaviour
        ).params.mutable_params.tx_hash = "test"
        self.behaviour.act_wrapper()

        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ),
            contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
            response_kwargs=response_kwargs,
        )

        if (
            response_kwargs["performative"]
            == ContractApiMessage.Performative.RAW_TRANSACTION
        ):
            self.mock_signing_request(
                request_kwargs=dict(
                    performative=SigningMessage.Performative.SIGN_TRANSACTION
                ),
                response_kwargs=dict(
                    performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                    signed_transaction=SignedTransaction(ledger_id="ethereum", body={}),
                ),
            )
            self.mock_ledger_api_request(
                request_kwargs=dict(
                    performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION
                ),
                response_kwargs=dict(
                    performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                    transaction_digest=TransactionDigest(
                        ledger_id="ethereum", body="tx_hash"
                    ),
                ),
            )

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == ValidateTransactionBehaviour.auto_behaviour_id()
        )
        assert (
            cast(
                ValidateTransactionBehaviour, self.behaviour.current_behaviour
            ).params.mutable_params.tx_hash
            == ""
        )

    def test_sender_act_tx_data_contains_tx_digest(self) -> None:
        """Test finalize behaviour."""

        max_priority_fee_per_gas: Optional[int] = None

        retries = 1
        participants = (
            self.skill.skill_context.agent_address,
            "a_1" + "-" * 39,
            "a_2" + "-" * 39,
        )
        kwargs = dict(
            safe_contract_address="safe_contract_address",
            participants=participants,
            participant_to_signature={},
            most_voted_tx_hash=hash_payload_to_hex(
                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                1,
                1,
                "0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215",
                b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
            ),
            nonce=None,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            keepers=retries.to_bytes(32, "big").hex()
            + self.skill.skill_context.agent_address,
        )

        db = AbciAppDB(setup_data=AbciAppDB.data_to_lists(kwargs))

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(db),
        )

        response_kwargs = dict(
            performative=ContractApiMessage.Performative.RAW_TRANSACTION,
            callable="get_deploy_transaction",
            raw_transaction=RawTransaction(
                ledger_id="ethereum",
                body={
                    "tx_hash": "0x3b",
                    "nonce": 0,
                    "maxFeePerGas": int(10e10),
                    "maxPriorityFeePerGas": int(10e10),
                },
            ),
        )

        # mock the returned tx_data
        return_value = dict(
            status=VerificationStatus.PENDING,
            keepers=deque(),
            keeper_retries=1,
            blacklisted_keepers=set(),
            tx_digest="dummy_tx_digest",
        )

        current_behaviour = cast(
            TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
        )
        current_behaviour._get_tx_data = mock_yield_and_return(return_value)  # type: ignore

        self.behaviour.act_wrapper()
        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ),
            contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
            response_kwargs=response_kwargs,
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)

        current_behaviour = cast(
            ValidateTransactionBehaviour, self.behaviour.current_behaviour
        )
        current_behaviour_id = current_behaviour.behaviour_id
        expected_behaviour_id = ValidateTransactionBehaviour.auto_behaviour_id()
        assert current_behaviour_id == expected_behaviour_id

    def test_handle_late_messages(self) -> None:
        """Test `handle_late_messages.`"""
        participants = (self.skill.skill_context.agent_address, "a_1", "a_2")
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            most_voted_keeper_address="most_voted_keeper_address",
                            participants=participants,
                            keepers="keepers",
                        )
                    ),
                )
            ),
        )
        self.behaviour.current_behaviour = cast(
            BaseBehaviour, self.behaviour.current_behaviour
        )
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )

        message = ContractApiMessage(ContractApiMessage.Performative.RAW_MESSAGE)  # type: ignore
        self.behaviour.current_behaviour.handle_late_messages(
            self.behaviour.current_behaviour.behaviour_id, message
        )
        assert cast(
            TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
        ).params.mutable_params.late_messages == [message]

        with mock.patch.object(self.behaviour.context.logger, "warning") as mock_info:
            self.behaviour.current_behaviour.handle_late_messages(
                "other_behaviour_id", message
            )
            mock_info.assert_called_with(
                f"No callback defined for request with nonce: {message.dialogue_reference[0]}, "
                "arriving for behaviour: other_behaviour_id"
            )
            message = MagicMock()
            self.behaviour.current_behaviour.handle_late_messages(
                self.behaviour.current_behaviour.behaviour_id, message
            )
            mock_info.assert_called_with(
                f"No callback defined for request with nonce: {message.dialogue_reference[0]}, "
                f"arriving for behaviour: {FinalizeBehaviour.auto_behaviour_id()}"
            )


class TestValidateTransactionBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test ValidateTransactionBehaviour."""

    def _fast_forward(self) -> None:
        """Fast-forward to relevant behaviour."""
        participants = (self.skill.skill_context.agent_address, "a_1", "a_2")
        most_voted_keeper_address = self.skill.skill_context.agent_address
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=ValidateTransactionBehaviour.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            safe_contract_address="safe_contract_address",
                            tx_hashes_history="t" * 66,
                            final_tx_hash="dummy_hash",
                            participants=participants,
                            most_voted_keeper_address=most_voted_keeper_address,
                            participant_to_signature={},
                            most_voted_tx_hash=hash_payload_to_hex(
                                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                                1,
                                1,
                                "0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215",
                                b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                            ),
                            max_priority_fee_per_gas=int(10e10),
                        )
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == ValidateTransactionBehaviour.auto_behaviour_id()
        )

    def test_validate_transaction_safe_behaviour(
        self,
    ) -> None:
        """Test ValidateTransactionBehaviour."""
        self._fast_forward()
        self.behaviour.act_wrapper()
        self.mock_ledger_api_request(
            request_kwargs=dict(
                performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT
            ),
            response_kwargs=dict(
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=TransactionReceipt(
                    ledger_id="ethereum", receipt={"status": 1}, transaction={}
                ),
            ),
        )
        self.mock_contract_api_request(
            request_kwargs=dict(performative=ContractApiMessage.Performative.GET_STATE),
            contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.STATE,
                callable="get_deploy_transaction",
                state=TrState(ledger_id="ethereum", body={"verified": True}),
            ),
        )
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert (
            behaviour.behaviour_id
            == make_degenerate_behaviour(
                FinishedTransactionSubmissionRound
            ).auto_behaviour_id()
        )

    def test_validate_transaction_safe_behaviour_no_tx_sent(
        self,
    ) -> None:
        """Test ValidateTransactionBehaviour when tx cannot be sent."""
        self._fast_forward()

        with mock.patch.object(self.behaviour.context.logger, "error") as mock_logger:
            self.behaviour.act_wrapper()
            self.mock_ledger_api_request(
                request_kwargs=dict(
                    performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                ),
                response_kwargs=dict(
                    performative=LedgerApiMessage.Performative.ERROR,
                    code=1,
                ),
            )
            behaviour = cast(
                TransactionSettlementBaseBehaviour,
                self.behaviour.current_behaviour,
            )
            latest_tx_hash = behaviour.synchronized_data.tx_hashes_history[-1]
            mock_logger.assert_any_call(f"tx {latest_tx_hash} receipt check timed out!")


class TestCheckTransactionHistoryBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test CheckTransactionHistoryBehaviour."""

    def _fast_forward(self, hashes_history: str) -> None:
        """Fast-forward to relevant behaviour."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=CheckTransactionHistoryBehaviour.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            safe_contract_address="safe_contract_address",
                            participants=(
                                self.skill.skill_context.agent_address,
                                "a_1",
                                "a_2",
                            ),
                            participant_to_signature={},
                            most_voted_tx_hash="b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d900000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000002625a000x77E9b2EF921253A171Fa0CB9ba80558648Ff7215b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                            tx_hashes_history=hashes_history,
                        )
                    ),
                )
            ),
        )
        assert (
            cast(BaseBehaviour, self.behaviour.current_behaviour).behaviour_id
            == CheckTransactionHistoryBehaviour.auto_behaviour_id()
        )

    @pytest.mark.parametrize(
        "verified, status, hashes_history, revert_reason",
        (
            (False, -1, "0x" + "t" * 64, "test"),
            (False, 0, "", "test"),
            (False, 0, "0x" + "t" * 64, "test"),
            (False, 0, "0x" + "t" * 64, "GS026"),
            (True, 1, "0x" + "t" * 64, "test"),
        ),
    )
    def test_check_tx_history_behaviour(
        self,
        verified: bool,
        status: int,
        hashes_history: str,
        revert_reason: str,
    ) -> None:
        """Test CheckTransactionHistoryBehaviour."""
        self._fast_forward(hashes_history)
        self.behaviour.act_wrapper()

        if hashes_history:
            self.mock_contract_api_request(
                request_kwargs=dict(
                    performative=ContractApiMessage.Performative.GET_STATE
                ),
                contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
                response_kwargs=dict(
                    performative=ContractApiMessage.Performative.STATE,
                    callable="verify_tx",
                    state=TrState(
                        ledger_id="ethereum",
                        body={
                            "verified": verified,
                            "status": status,
                            "transaction": {},
                        },
                    ),
                ),
            )

            if not verified and status != -1:
                self.mock_contract_api_request(
                    request_kwargs=dict(
                        performative=ContractApiMessage.Performative.GET_STATE
                    ),
                    contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
                    response_kwargs=dict(
                        performative=ContractApiMessage.Performative.STATE,
                        callable="revert_reason",
                        state=TrState(
                            ledger_id="ethereum", body={"revert_reason": revert_reason}
                        ),
                    ),
                )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert (
            behaviour.behaviour_id
            == make_degenerate_behaviour(
                FinishedTransactionSubmissionRound
            ).auto_behaviour_id()
        )


class TestCheckLateTxHashesBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test CheckLateTxHashesBehaviour."""

    def _fast_forward(self, late_arriving_tx_hashes: Dict[str, str]) -> None:
        """Fast-forward to relevant behaviour."""

        agent_address = self.skill.skill_context.agent_address
        kwargs = dict(
            safe_contract_address="safe_contract_address",
            participants=(agent_address, "a_1", "a_2"),
            participant_to_signature={},
            most_voted_tx_hash="",
            late_arriving_tx_hashes=late_arriving_tx_hashes,
        )
        abci_app_db = AbciAppDB(setup_data=AbciAppDB.data_to_lists(kwargs))

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=CheckLateTxHashesBehaviour.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(abci_app_db),
        )

        current_behaviour = self.behaviour.current_behaviour
        current_behaviour_id = cast(BaseBehaviour, current_behaviour).behaviour_id
        assert current_behaviour_id == CheckLateTxHashesBehaviour.auto_behaviour_id()

    def test_check_tx_history_behaviour(self) -> None:
        """Test CheckTransactionHistoryBehaviour."""
        self._fast_forward(late_arriving_tx_hashes={})
        self.behaviour.act_wrapper()
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        next_degen_behaviour = make_degenerate_behaviour(
            FinishedTransactionSubmissionRound
        )
        assert behaviour.behaviour_id == next_degen_behaviour.auto_behaviour_id()


class TestSynchronizeLateMessagesBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test `SynchronizeLateMessagesBehaviour`"""

    def _check_behaviour_id(
        self, expected: Type[TransactionSettlementBaseBehaviour]
    ) -> None:
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == expected.auto_behaviour_id()

    @pytest.mark.parametrize("late_messages", ([], [MagicMock, MagicMock]))
    def test_async_act(self, late_messages: List[MagicMock]) -> None:
        """Test `async_act`"""
        cast(
            TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
        ).params.mutable_params.late_messages = late_messages

        participants = (self.skill.skill_context.agent_address, "a_1", "a_2")
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=SynchronizeLateMessagesBehaviour.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(
                    setup_data=dict(
                        participants=[participants],
                        participant_to_signature=[{}],
                        safe_contract_address=["safe_contract_address"],
                        most_voted_tx_hash=[
                            hash_payload_to_hex(
                                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                                1,
                                1,
                                "0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215",
                                b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9"
                                b"b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                            )
                        ],
                    ),
                )
            ),
        )
        self._check_behaviour_id(SynchronizeLateMessagesBehaviour)  # type: ignore

        if not late_messages:
            self.behaviour.act_wrapper()
            self.mock_a2a_transaction()
            self._test_done_flag_set()
            self.end_round(TransactionSettlementEvent.DONE)
            self._check_behaviour_id(CheckLateTxHashesBehaviour)  # type: ignore

        else:

            def _dummy_get_tx_data(
                _current_message: ContractApiMessage,
                _use_flashbots: bool,
            ) -> Generator[None, None, TxDataType]:
                yield
                return {
                    "status": VerificationStatus.PENDING,
                    "tx_digest": "test",
                }

            cast(
                TransactionSettlementBaseBehaviour, self.behaviour.current_behaviour
            )._get_tx_data = _dummy_get_tx_data  # type: ignore
            for _ in range(len(late_messages)):
                self.behaviour.act_wrapper()


class TestResetBehaviour(TransactionSettlementFSMBehaviourBaseCase):
    """Test the reset behaviour."""

    behaviour_class = ResetBehaviour
    next_behaviour_class = RandomnessTransactionSubmissionBehaviour

    def test_reset_behaviour(
        self,
    ) -> None:
        """Test reset behaviour."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=TransactionSettlementSynchronizedSata(
                AbciAppDB(setup_data=dict(estimate=[1.0])),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        self.behaviour.context.params.__dict__["reset_pause_duration"] = 0.1
        self.behaviour.act_wrapper()
        time.sleep(0.3)
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(TransactionSettlementEvent.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.auto_behaviour_id()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Environment variable helpers."""


import json
import os
from pathlib import Path
from typing import Dict

from dotenv import load_dotenv


def load_json(file: Path, serialize: bool = False) -> None:
    """Load json."""
    env_vars: Dict = json.load(file.open(mode="r", encoding="utf-8"))
    if serialize:
        for key, val in env_vars.items():
            if isinstance(val, str):
                continue  # pragma: nocover
            env_vars[key] = json.dumps(val)
    os.environ.update(env_vars)


def load_env_file(file: Path, serialize_json: bool = False) -> None:
    """Load env file."""

    if file.name.endswith(".json"):
        load_json(file=file, serialize=serialize_json)
    else:
        load_dotenv(dotenv_path=file)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the handlers.py module of the skill."""

# pylint: skip-file

import json
import logging
from dataclasses import asdict
from typing import Any, Dict, cast
from unittest import mock
from unittest.mock import MagicMock

import pytest
from _pytest.logging import LogCaptureFixture
from aea.configurations.data_types import PublicId
from aea.protocols.base import Message

from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import (
    CheckTxType,
    CheckTxTypeEnum,
    ConsensusParams,
    Evidences,
    Header,
    LastCommitInfo,
    Timestamp,
    ValidatorUpdates,
)
from packages.valory.protocols.http import HttpMessage
from packages.valory.protocols.tendermint import TendermintMessage
from packages.valory.skills.abstract_round_abci import handlers
from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AddBlockError,
    ERROR_CODE,
    OK_CODE,
    SignatureNotValidError,
)
from packages.valory.skills.abstract_round_abci.dialogues import (
    AbciDialogue,
    AbciDialogues,
    TendermintDialogue,
    TendermintDialogues,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    ABCIRoundHandler,
    AbstractResponseHandler,
    TendermintHandler,
    Transaction,
    exception_to_info_msg,
)
from packages.valory.skills.abstract_round_abci.models import TendermintRecoveryParams
from packages.valory.skills.abstract_round_abci.test_tools.rounds import DummyRound


def test_exception_to_info_msg() -> None:
    """Test 'exception_to_info_msg' helper function."""
    exception = Exception("exception message")
    expected_string = f"{exception.__class__.__name__}: {str(exception)}"
    actual_string = exception_to_info_msg(exception)
    assert expected_string == actual_string


class TestABCIRoundHandler:
    """Test 'ABCIRoundHandler'."""

    def setup(self) -> None:
        """Set up the tests."""
        self.context = MagicMock(skill_id=PublicId.from_str("dummy/skill:0.1.0"))
        self.dialogues = AbciDialogues(name="", skill_context=self.context)
        self.handler = ABCIRoundHandler(name="", skill_context=self.context)
        self.context.state.round_sequence.height = 0
        self.context.state.round_sequence.root_hash = b"root_hash"

    def test_info(self) -> None:
        """Test the 'info' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_INFO,
            version="",
            block_version=0,
            p2p_version=0,
        )
        response = self.handler.info(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_INFO

    @pytest.mark.parametrize("app_hash", (b"", b"test"))
    def test_init_chain(self, app_hash: bytes) -> None:
        """Test the 'init_chain' handler method."""
        time = Timestamp(0, 0)
        consensus_params = ConsensusParams(*(mock.MagicMock() for _ in range(4)))
        validators = ValidatorUpdates(mock.MagicMock())
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
            time=time,
            chain_id="test_chain_id",
            consensus_params=consensus_params,
            validators=validators,
            app_state_bytes=b"",
            initial_height=10,
        )
        self.context.state.round_sequence.last_round_transition_root_hash = app_hash
        response = self.handler.init_chain(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_INIT_CHAIN
        assert response.validators == ValidatorUpdates([])
        assert response.app_hash == app_hash

    def test_begin_block(self) -> None:
        """Test the 'begin_block' handler method."""
        header = Header(*(MagicMock() for _ in range(14)))
        last_commit_info = LastCommitInfo(*(MagicMock() for _ in range(2)))
        byzantine_validators = Evidences(MagicMock())
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,
            hash=b"",
            header=header,
            last_commit_info=last_commit_info,
            byzantine_validators=byzantine_validators,
        )
        response = self.handler.begin_block(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_BEGIN_BLOCK

    @mock.patch.object(handlers, "Transaction")
    def test_check_tx(self, *_: Any) -> None:
        """Test the 'check_tx' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,
            tx=b"",
            type=CheckTxType(CheckTxTypeEnum.NEW),
        )
        response = self.handler.check_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_CHECK_TX
        assert response.code == OK_CODE

    @mock.patch.object(
        Transaction,
        "decode",
        side_effect=SignatureNotValidError,
    )
    def test_check_tx_negative(self, *_: Any) -> None:
        """Test the 'check_tx' handler method, negative case."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,
            tx=b"",
            type=CheckTxType(CheckTxTypeEnum.NEW),
        )
        response = self.handler.check_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_CHECK_TX
        assert response.code == ERROR_CODE

    @mock.patch.object(handlers, "Transaction")
    def test_deliver_tx(self, *_: Any) -> None:
        """Test the 'deliver_tx' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
            tx=b"",
        )
        response = self.handler.deliver_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_DELIVER_TX
        assert response.code == OK_CODE

    @mock.patch.object(
        Transaction,
        "decode",
        side_effect=SignatureNotValidError,
    )
    def test_deliver_tx_negative(self, *_: Any) -> None:
        """Test the 'deliver_tx' handler method, negative case."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
            tx=b"",
        )
        response = self.handler.deliver_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_DELIVER_TX
        assert response.code == ERROR_CODE

    @pytest.mark.parametrize("request_height", tuple(range(3)))
    def test_end_block(self, request_height: int) -> None:
        """Test the 'end_block' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_END_BLOCK,
            height=request_height,
        )
        assert isinstance(message, AbciMessage)
        assert isinstance(dialogue, AbciDialogue)
        assert message.height == request_height
        assert self.context.state.round_sequence.tm_height != request_height
        response = self.handler.end_block(message, dialogue)
        assert response.performative == AbciMessage.Performative.RESPONSE_END_BLOCK
        assert self.context.state.round_sequence.tm_height == request_height

    def test_commit(self) -> None:
        """Test the 'commit' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_COMMIT,
        )
        response = self.handler.commit(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_COMMIT

    def test_commit_negative(self) -> None:
        """Test the 'commit' handler method, negative case."""
        self.context.state.round_sequence.commit.side_effect = AddBlockError()
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_COMMIT,
        )
        with pytest.raises(AddBlockError):
            self.handler.commit(
                cast(AbciMessage, message), cast(AbciDialogue, dialogue)
            )


class ConcreteResponseHandler(AbstractResponseHandler):
    """A concrete response handler for testing purposes."""

    SUPPORTED_PROTOCOL = HttpMessage.protocol_id
    allowed_response_performatives = frozenset({HttpMessage.Performative.RESPONSE})


class TestAbstractResponseHandler:
    """Test 'AbstractResponseHandler'."""

    def setup(self) -> None:
        """Set up the tests."""
        self.context = MagicMock()
        self.handler = ConcreteResponseHandler(name="", skill_context=self.context)

    def test_handle(self) -> None:
        """Test the 'handle' method."""
        callback = MagicMock()
        request_reference = "reference"
        self.context.requests.request_id_to_callback = {}
        self.context.requests.request_id_to_callback[request_reference] = callback
        with mock.patch.object(
            self.handler, "_recover_protocol_dialogues"
        ) as mock_dialogues_fn:
            mock_dialogue = MagicMock()
            mock_dialogue.dialogue_label.dialogue_reference = (request_reference, "")
            mock_dialogues = MagicMock()
            mock_dialogues.update = MagicMock(return_value=mock_dialogue)
            mock_dialogues_fn.return_value = mock_dialogues
            mock_message = MagicMock(performative=HttpMessage.Performative.RESPONSE)
            self.handler.handle(mock_message)
        callback.assert_called()

    @mock.patch.object(
        AbstractResponseHandler, "_recover_protocol_dialogues", return_value=None
    )
    def test_handle_negative_cannot_recover_dialogues(self, *_: Any) -> None:
        """Test the 'handle' method, negative case (cannot recover dialogues)."""
        self.handler.handle(MagicMock())

    @mock.patch.object(AbstractResponseHandler, "_recover_protocol_dialogues")
    def test_handle_negative_cannot_update_dialogues(
        self, mock_dialogues_fn: Any
    ) -> None:
        """Test the 'handle' method, negative case (cannot update dialogues)."""
        mock_dialogues = MagicMock(update=MagicMock(return_value=None))
        mock_dialogues_fn.return_value = mock_dialogues
        self.handler.handle(MagicMock())

    def test_handle_negative_performative_not_allowed(self) -> None:
        """Test the 'handle' method, negative case (performative not allowed)."""
        self.handler.handle(MagicMock())

    def test_handle_negative_cannot_find_callback(self) -> None:
        """Test the 'handle' method, negative case (cannot find callback)."""
        self.context.requests.request_id_to_callback = {}
        with pytest.raises(
            ABCIAppInternalError, match="No callback defined for request with nonce: "
        ):
            self.handler.handle(
                MagicMock(performative=HttpMessage.Performative.RESPONSE)
            )


class TestTendermintHandler:
    """Test Tendermint Handler"""

    def setup(self) -> None:
        """Set up the tests."""
        self.agent_name = "Alice"
        self.context = MagicMock(skill_id=PublicId.from_str("dummy/skill:0.1.0"))
        other_agents = ["Alice", "Bob", "Charlie"]
        self.context.state = MagicMock(acn_container=lambda: other_agents)
        self.handler = TendermintHandler(name="dummy", skill_context=self.context)
        self.handler.context.logger = logging.getLogger()
        self.dialogues = TendermintDialogues(name="dummy", skill_context=self.context)

    @property
    def dummy_validator_config(self) -> Dict[str, Dict[str, str]]:
        """Dummy validator config"""
        return {
            self.agent_name: {
                "hostname": "localhost",
                "address": "address",
                "pub_key": "pub_key",
                "peer_id": "peer_id",
            }
        }

    @staticmethod
    def make_error_message() -> TendermintMessage:
        """Make dummy error message"""
        performative = TendermintMessage.Performative.ERROR
        error_code = TendermintMessage.ErrorCode.INVALID_REQUEST
        error_msg, error_data = "", {}  # type: ignore
        message = TendermintMessage(
            performative,  # type: ignore
            error_code=error_code,
            error_msg=error_msg,
            error_data=error_data,
        )
        message.sender = "Alice"
        return message

    # pre-condition checks
    def test_handle_unidentified_tendermint_dialogue(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test unidentified tendermint dialogue"""
        message = Message()
        with mock.patch.object(self.handler.dialogues, "update", return_value=None):
            self.handler.handle(message)
            log_message = self.handler.LogMessages.unidentified_dialogue.value
            assert log_message in caplog.text

    def test_handle_no_addresses_retrieved_yet(self, caplog: LogCaptureFixture) -> None:
        """Test handle request no registered addresses"""
        performative = TendermintMessage.Performative.GET_GENESIS_INFO
        message = TendermintMessage(performative)  # type: ignore
        message.sender = "Alice"
        self.handler.initial_tm_configs = {}
        self.handler.handle(message)
        log_message = self.handler.LogMessages.no_addresses_retrieved_yet.value
        assert log_message in caplog.text
        log_message = self.handler.LogMessages.sending_error_response.value
        assert log_message in caplog.text

    def test_handle_not_in_registered_addresses(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test handle response sender not in registered addresses"""
        performative = TendermintMessage.Performative.GENESIS_INFO
        message = TendermintMessage(performative, info="info")  # type: ignore
        message.sender = "NotAlice"
        self.handler.handle(message)
        log_message = self.handler.LogMessages.not_in_registered_addresses.value
        assert log_message in caplog.text

    # request
    def test_handle_get_genesis_info(self, caplog: LogCaptureFixture) -> None:
        """Test handle request for genesis info"""
        performative = TendermintMessage.Performative.GET_GENESIS_INFO
        message = TendermintMessage(performative)  # type: ignore
        self.context.agent_address = message.sender = self.agent_name
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.sending_request_response.value
        assert log_message in caplog.text

    # response
    def test_handle_response_invalid_addresses(self, caplog: LogCaptureFixture) -> None:
        """Test handle response for genesis info with invalid address."""
        validator_config = self.dummy_validator_config
        validator_config[self.agent_name]["hostname"] = "random"
        performative = TendermintMessage.Performative.GENESIS_INFO
        info = json.dumps(validator_config[self.agent_name])
        message = TendermintMessage(performative, info=info)  # type: ignore
        self.context.agent_address = message.sender = self.agent_name
        self.handler.initial_tm_configs = validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.failed_to_parse_address.value
        assert log_message in caplog.text

    def test_handle_genesis_info(self, caplog: LogCaptureFixture) -> None:
        """Test handle response for genesis info with valid address"""
        performative = TendermintMessage.Performative.GENESIS_INFO
        info = json.dumps(self.dummy_validator_config[self.agent_name])
        message = TendermintMessage(performative, info=info)  # type: ignore
        self.context.agent_address = message.sender = self.agent_name
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.collected_config_info.value
        assert log_message in caplog.text

    @pytest.mark.parametrize("registered", (True, False))
    @pytest.mark.parametrize(
        "performative",
        (
            TendermintMessage.Performative.RECOVERY_PARAMS,
            TendermintMessage.Performative.GET_RECOVERY_PARAMS,
        ),
    )
    def test_recovery_params(
        self,
        registered: bool,
        performative: TendermintMessage.Performative,
        caplog: LogCaptureFixture,
    ) -> None:
        """Test handle response for recovery parameters."""
        if not registered:
            self.agent_name = "not-registered"

        if performative == TendermintMessage.Performative.GET_RECOVERY_PARAMS:
            self.context.state.tm_recovery_params = TendermintRecoveryParams(
                "DummyRound"
            )
            message = TendermintMessage(performative)  # type: ignore
            log_message = self.handler.LogMessages.sending_request_response.value
        elif performative == TendermintMessage.Performative.RECOVERY_PARAMS:
            params = json.dumps(
                asdict(TendermintRecoveryParams(DummyRound.auto_round_id()))
            )
            message = TendermintMessage(performative, params=params)  # type: ignore
            log_message = self.handler.LogMessages.collected_params.value
        else:
            raise AssertionError(
                f"Invalid performative {performative} for `test_recovery_params`."
            )

        self.context.agent_address = message.sender = self.agent_name
        tm_configs = {self.agent_name: {"dummy": "value"}} if registered else {}

        self.handler.initial_tm_configs = tm_configs
        self.handler.handle(message)

        if not registered:
            log_message = self.handler.LogMessages.not_in_registered_addresses.value

        assert log_message in caplog.text

    @pytest.mark.parametrize(
        "side_effect, expected_exception",
        (
            (
                json.decoder.JSONDecodeError("", "", 0),
                ": line 1 column 1 (char 0)",
            ),
            (
                {"not a dict"},
                "argument after ** must be a mapping, not str",
            ),
        ),
    )
    def test_recovery_params_error(
        self,
        side_effect: Any,
        expected_exception: str,
        caplog: LogCaptureFixture,
    ) -> None:
        """Test handle response for recovery parameters."""
        message = TendermintMessage(
            TendermintMessage.Performative.RECOVERY_PARAMS, params=MagicMock()  # type: ignore
        )

        self.context.agent_address = message.sender = self.agent_name
        tm_configs = {self.agent_name: {"dummy": "value"}}
        self.handler.initial_tm_configs = tm_configs
        with mock.patch.object(json, "loads", side_effect=side_effect):
            self.handler.handle(message)

        log_message = self.handler.LogMessages.failed_to_parse_params.value
        assert log_message in caplog.text
        assert expected_exception in caplog.text

    # error
    def test_handle_error(self, caplog: LogCaptureFixture) -> None:
        """Test handle error"""
        message = self.make_error_message()
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.received_error_response.value
        assert log_message in caplog.text

    def test_handle_error_no_target_message_retrieved(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test handle error no target message retrieved"""
        message, nonce = self.make_error_message(), "0"
        dialogue = TendermintDialogue(mock.Mock(), "Bob", mock.Mock())
        dialogue.dialogue_label.dialogue_reference = nonce, "stub"
        self.handler.dialogues.update = lambda _: dialogue  # type: ignore
        callback = lambda *args, **kwargs: None  # noqa: E731
        self.context.requests.request_id_to_callback = {nonce: callback}
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = (
            self.handler.LogMessages.received_error_without_target_message.value
        )
        assert log_message in caplog.text

    # performative
    def test_handle_performative_not_recognized(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test performative no recognized"""
        message = self.make_error_message()
        message._slots.performative = MagicMock(value="wacky")
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.performative_not_recognized.value
        assert log_message in caplog.text

    def test_sender_not_in_registered_addresses(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test sender not in registered addresses."""

        performative = TendermintMessage.Performative.GET_GENESIS_INFO
        message = TendermintMessage(performative)  # type: ignore
        self.context.agent_address = message.sender = "dummy"
        self.handler.initial_tm_configs = self.dummy_validator_config
        self.handler.handle(message)
        log_message = self.handler.LogMessages.not_in_registered_addresses.value
        assert log_message in caplog.text



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for docstring generator."""


import os
from pathlib import Path
from typing import Tuple
from unittest import mock

from aea.configurations.constants import PACKAGES

from packages.valory.skills.hello_world_abci import rounds

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestDocstrings(BaseCliTest):
    """Test `autonomy analyse docstrings`."""

    rounds_file_original: Path
    rounds_file_temp: Path
    skill_name: str = "hello_world_abci"
    cli_options: Tuple[str, ...] = ("analyse", "docstrings")
    rounds_file = Path(PACKAGES, "valory", "skills", skill_name, "rounds.py")

    def setup(
        self,
    ) -> None:
        """Setup test method."""
        super().setup()

        os.chdir(self.t)

        packages_dir = self.t / PACKAGES
        packages_dir.mkdir()

        (packages_dir / "valory").mkdir()
        (packages_dir / "valory" / "skills").mkdir()
        (packages_dir / "valory" / "skills" / self.skill_name).mkdir()

        self.rounds_file_original = Path(*ROOT_DIR.parts, *self.rounds_file.parts)
        self.rounds_file_temp = Path(*self.t.parts, *self.rounds_file.parts)

        self.rounds_file_temp.write_text(self.rounds_file_original.read_text())

    def _get_expected_output(
        self,
    ) -> str:
        """Return expected output."""
        expcted_output = ""
        for file in sorted(self.t.glob("packages/*/skills/*/rounds.py")):
            expcted_output += f"Processing: {file.relative_to(self.t)}\n"
        return expcted_output

    def _corrupt_round_file(
        self,
    ) -> None:
        """Corrupt a round file."""

        string_to_replace = """4. ResetAndPauseRound
            - done: 1.
            - no majority: 0.
            - reset timeout: 0.\n"""

        rounds_file_content = self.rounds_file_original.read_text()
        rounds_file_content = rounds_file_content.replace(string_to_replace, "")
        self.rounds_file_temp.write_text(rounds_file_content)

    def test_docstring_check(
        self,
    ) -> None:
        """Test after check"""

        self.run_cli()
        result = self.run_cli()

        assert result.exit_code == 0, result.output
        assert f"Processing skill {self.skill_name} with author valory" in result.output
        assert "No update needed." in result.output

    def test_docstring_check_fail(
        self,
    ) -> None:
        """Test after check"""

        self.run_cli()
        self._corrupt_round_file()

        new_callable = mock.PropertyMock(return_value=str(self.rounds_file_temp))
        with mock.patch.object(rounds, "__file__", new_callable=new_callable):
            result = self.run_cli()

        assert result.exit_code == 1, result.output
        assert "Error: Following files needs updating" in result.output

    def test_fix_docstring(
        self,
    ) -> None:
        """Test after check"""

        self._corrupt_round_file()
        assert (
            self.rounds_file_temp.read_text() != self.rounds_file_original.read_text()
        )

        new_callable = mock.PropertyMock(return_value=str(self.rounds_file_temp))
        with mock.patch.object(rounds, "__file__", new_callable=new_callable):
            result = self.run_cli(("--update",))
        assert result.exit_code == 0, result.output
        assert "Updated following files" in result.output
        assert (
            self.rounds_file_temp.read_text() == self.rounds_file_original.read_text()
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/registration_abci skill's rounds."""

# pylint: skip-file

import hashlib
import logging  # noqa: F401
from collections import deque
from typing import (
    Any,
    Deque,
    Dict,
    FrozenSet,
    List,
    Mapping,
    Optional,
    Type,
    Union,
    cast,
)
from unittest import mock

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciAppDB,
)
from packages.valory.skills.abstract_round_abci.base import (
    BaseSynchronizedData as SynchronizedData,
)
from packages.valory.skills.abstract_round_abci.base import (
    BaseTxPayload,
    CollectSameUntilThresholdRound,
    CollectionRound,
    MAX_INT_256,
    TransactionNotValidError,
    VotingRound,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseCollectDifferentUntilThresholdRoundTest,
    BaseCollectNonEmptyUntilThresholdRound,
    BaseCollectSameUntilThresholdRoundTest,
    BaseOnlyKeeperSendsRoundTest,
    BaseVotingRoundTest,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
)
from packages.valory.skills.transaction_settlement_abci.payloads import (
    CheckTransactionHistoryPayload,
    FinalizationTxPayload,
    RandomnessPayload,
    ResetPayload,
    SelectKeeperPayload,
    SignaturePayload,
    SynchronizeLateMessagesPayload,
    ValidatePayload,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    CheckTransactionHistoryRound,
    CollectSignatureRound,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    Event as TransactionSettlementEvent,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    FinalizationRound,
    ResetRound,
    SelectKeeperTransactionSubmissionARound,
    SelectKeeperTransactionSubmissionBAfterTimeoutRound,
    SelectKeeperTransactionSubmissionBRound,
    SynchronizeLateMessagesRound,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TransactionSettlementSynchronizedSata,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    TX_HASH_LENGTH,
    ValidateTransactionRound,
)


MAX_PARTICIPANTS: int = 4
RANDOMNESS: str = "d1c29dce46f979f9748210d24bce4eae8be91272f5ca1a6aea2832d3dd676f51"
DUMMY_RANDOMNESS = hashlib.sha256("hash".encode() + str(0).encode()).hexdigest()


def get_participants() -> FrozenSet[str]:
    """Participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


def get_participant_to_randomness(
    participants: FrozenSet[str], round_id: int
) -> Dict[str, RandomnessPayload]:
    """participant_to_randomness"""
    return {
        participant: RandomnessPayload(
            sender=participant,
            round_id=round_id,
            randomness=RANDOMNESS,
        )
        for participant in participants
    }


def get_most_voted_randomness() -> str:
    """most_voted_randomness"""
    return RANDOMNESS


def get_participant_to_selection(
    participants: FrozenSet[str],
    keepers: str,
) -> Dict[str, SelectKeeperPayload]:
    """participant_to_selection"""
    return {
        participant: SelectKeeperPayload(sender=participant, keepers=keepers)
        for participant in participants
    }


def get_participant_to_period_count(
    participants: FrozenSet[str], period_count: int
) -> Dict[str, ResetPayload]:
    """participant_to_selection"""
    return {
        participant: ResetPayload(sender=participant, period_count=period_count)
        for participant in participants
    }


def get_safe_contract_address() -> str:
    """safe_contract_address"""
    return "0x6f6ab56aca12"


def get_participant_to_votes(
    participants: FrozenSet[str], vote: Optional[bool] = True
) -> Dict[str, ValidatePayload]:
    """participant_to_votes"""
    return {
        participant: ValidatePayload(sender=participant, vote=vote)
        for participant in participants
    }


def get_participant_to_votes_serialized(
    participants: FrozenSet[str], vote: Optional[bool] = True
) -> Dict[str, Dict[str, Any]]:
    """participant_to_votes"""
    return CollectionRound.serialize_collection(
        get_participant_to_votes(participants, vote)
    )


def get_most_voted_tx_hash() -> str:
    """most_voted_tx_hash"""
    return "tx_hash"


def get_participant_to_signature(
    participants: FrozenSet[str],
) -> Dict[str, SignaturePayload]:
    """participant_to_signature"""
    return {
        participant: SignaturePayload(sender=participant, signature="signature")
        for participant in participants
    }


def get_final_tx_hash() -> str:
    """final_tx_hash"""
    return "tx_hash"


def get_participant_to_check(
    participants: FrozenSet[str],
    status: str,
    tx_hash: str,
) -> Dict[str, CheckTransactionHistoryPayload]:
    """Get participants to check"""
    return {
        participant: CheckTransactionHistoryPayload(
            sender=participant,
            verified_res=status + tx_hash,
        )
        for participant in participants
    }


def get_participant_to_late_arriving_tx_hashes(
    participants: FrozenSet[str],
) -> Dict[str, SynchronizeLateMessagesPayload]:
    """participant_to_selection"""
    return {
        participant: SynchronizeLateMessagesPayload(
            sender=participant, tx_hashes="1" * TX_HASH_LENGTH + "2" * TX_HASH_LENGTH
        )
        for participant in participants
    }


def get_late_arriving_tx_hashes_deserialized() -> Dict[str, List[str]]:
    """Get dummy late-arriving tx hashes."""
    # We want the tx hashes to have a size which can be divided by 64 to be able to parse it.
    # Otherwise, they are not valid.
    return {
        "sender": [
            "t" * TX_HASH_LENGTH,
            "e" * TX_HASH_LENGTH,
            "s" * TX_HASH_LENGTH,
            "t" * TX_HASH_LENGTH,
        ]
    }


def get_late_arriving_tx_hashes_serialized() -> Dict[str, str]:
    """Get dummy late-arriving tx hashes."""
    # We want the tx hashes to have a size which can be divided by 64 to be able to parse it.
    # Otherwise, they are not valid.
    deserialized = get_late_arriving_tx_hashes_deserialized()
    return {sender: "".join(hash_) for sender, hash_ in deserialized.items()}


def get_keepers(keepers: Deque[str], retries: int = 1) -> str:
    """Get dummy keepers."""
    return retries.to_bytes(32, "big").hex() + "".join(keepers)


class BaseValidateRoundTest(BaseVotingRoundTest):
    """Test BaseValidateRound."""

    test_class: Type[VotingRound]
    test_payload: Type[ValidatePayload]

    def test_positive_votes(
        self,
    ) -> None:
        """Test ValidateRound."""

        self.synchronized_data.update(tx_hashes_history="t" * 66)

        test_round = self.test_class(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_voting_round_positive(
                test_round=test_round,
                round_payloads=get_participant_to_votes(self.participants),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    participant_to_votes=get_participant_to_votes_serialized(
                        self.participants
                    )
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_votes.keys()
                ],
                exit_event=self._event_class.DONE,
            )
        )

    def test_negative_votes(
        self,
    ) -> None:
        """Test ValidateRound."""

        test_round = self.test_class(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_voting_round_negative(
                test_round=test_round,
                round_payloads=get_participant_to_votes(self.participants, vote=False),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    participant_to_votes=get_participant_to_votes_serialized(
                        self.participants, vote=False
                    )
                ),
                synchronized_data_attr_checks=[],
                exit_event=self._event_class.NEGATIVE,
            )
        )

    def test_none_votes(
        self,
    ) -> None:
        """Test ValidateRound."""

        test_round = self.test_class(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_voting_round_none(
                test_round=test_round,
                round_payloads=get_participant_to_votes(self.participants, vote=None),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    participant_to_votes=get_participant_to_votes_serialized(
                        self.participants, vote=None
                    )
                ),
                synchronized_data_attr_checks=[],
                exit_event=self._event_class.NONE,
            )
        )


class BaseSelectKeeperRoundTest(BaseCollectSameUntilThresholdRoundTest):
    """Test SelectKeeperTransactionSubmissionARound"""

    test_class: Type[CollectSameUntilThresholdRound]
    test_payload: Type[BaseTxPayload]

    _synchronized_data_class = SynchronizedData

    @staticmethod
    def _participant_to_selection(
        participants: FrozenSet[str], keepers: str
    ) -> Mapping[str, BaseTxPayload]:
        """Get participant to selection"""
        return get_participant_to_selection(participants, keepers)

    def test_run(
        self,
        most_voted_payload: str = "keeper",
        keepers: str = "",
        exit_event: Optional[Any] = None,
    ) -> None:
        """Run tests."""
        test_round = self.test_class(
            synchronized_data=self.synchronized_data.update(
                keepers=keepers,
            ),
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=self._participant_to_selection(
                    self.participants, most_voted_payload
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                    participant_to_selection=CollectionRound.serialize_collection(
                        self._participant_to_selection(
                            self.participants, most_voted_payload
                        )
                    )
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_selection.keys()
                    if exit_event is None
                    else None
                ],
                most_voted_payload=most_voted_payload,
                exit_event=self._event_class.DONE if exit_event is None else exit_event,
            )
        )


class TestSelectKeeperTransactionSubmissionARound(BaseSelectKeeperRoundTest):
    """Test SelectKeeperTransactionSubmissionARound"""

    test_class = SelectKeeperTransactionSubmissionARound
    test_payload = SelectKeeperPayload
    _synchronized_data_class = TransactionSettlementSynchronizedSata
    _event_class = TransactionSettlementEvent

    @pytest.mark.parametrize(
        "most_voted_payload, keepers, exit_event",
        (
            (
                "incorrectly_serialized",
                "",
                TransactionSettlementEvent.INCORRECT_SERIALIZATION,
            ),
            (
                int(1).to_bytes(32, "big").hex() + "new_keeper" + "-" * 32,
                "",
                TransactionSettlementEvent.DONE,
            ),
        ),
    )
    def test_run(
        self,
        most_voted_payload: str,
        keepers: str,
        exit_event: TransactionSettlementEvent,
    ) -> None:
        """Run tests."""
        super().test_run(most_voted_payload, keepers, exit_event)


class TestSelectKeeperTransactionSubmissionBRound(
    TestSelectKeeperTransactionSubmissionARound
):
    """Test SelectKeeperTransactionSubmissionBRound."""

    test_class = SelectKeeperTransactionSubmissionBRound

    @pytest.mark.parametrize(
        "most_voted_payload, keepers, exit_event",
        (
            (
                int(1).to_bytes(32, "big").hex() + "new_keeper" + "-" * 32,
                "",
                TransactionSettlementEvent.DONE,
            ),
            (
                int(1).to_bytes(32, "big").hex() + "new_keeper" + "-" * 32,
                int(1).to_bytes(32, "big").hex()
                + "".join(
                    [keeper + "-" * 30 for keeper in ("test_keeper1", "test_keeper2")]
                ),
                TransactionSettlementEvent.DONE,
            ),
        ),
    )
    def test_run(
        self,
        most_voted_payload: str,
        keepers: str,
        exit_event: TransactionSettlementEvent,
    ) -> None:
        """Run tests."""
        super().test_run(most_voted_payload, keepers, exit_event)


class TestSelectKeeperTransactionSubmissionBAfterTimeoutRound(
    TestSelectKeeperTransactionSubmissionBRound
):
    """Test SelectKeeperTransactionSubmissionBAfterTimeoutRound."""

    test_class = SelectKeeperTransactionSubmissionBAfterTimeoutRound

    @mock.patch.object(
        TransactionSettlementSynchronizedSata,
        "keepers_threshold_exceeded",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        "keepers", (f"{int(1).to_bytes(32, 'big').hex()}keeper" + "-" * 36,)
    )
    @pytest.mark.parametrize(
        "attrs, threshold_exceeded, exit_event",
        (
            (
                {
                    "tx_hashes_history": "t" * 66,
                    "missed_messages": {f"keeper{'-' * 36}": 10},
                },
                True,
                # Since the threshold has been exceeded, we should return a `CHECK_HISTORY` event.
                TransactionSettlementEvent.CHECK_HISTORY,
            ),
            (
                {
                    "missed_messages": {f"keeper{'-' * 36}": 10},
                },
                True,
                TransactionSettlementEvent.CHECK_LATE_ARRIVING_MESSAGE,
            ),
            (
                {
                    "missed_messages": {f"keeper{'-' * 36}": 10},
                },
                False,
                TransactionSettlementEvent.DONE,
            ),
        ),
    )
    def test_run(
        self,
        threshold_exceeded_mock: mock.PropertyMock,
        keepers: str,
        attrs: Dict[str, Union[str, int]],
        threshold_exceeded: bool,
        exit_event: TransactionSettlementEvent,
    ) -> None:
        """Test `SelectKeeperTransactionSubmissionBAfterTimeoutRound`."""
        self.synchronized_data.update(participant_to_selection=dict.fromkeys(self.participants), **attrs)  # type: ignore
        threshold_exceeded_mock.return_value = threshold_exceeded
        most_voted_payload = int(1).to_bytes(32, "big").hex() + "new_keeper" + "-" * 32
        super().test_run(most_voted_payload, keepers, exit_event)
        initial_missed_messages = cast(Dict[str, int], (attrs["missed_messages"]))
        expected_missed_messages = {
            sender: missed + 1 for sender, missed in initial_missed_messages.items()
        }
        synchronized_data = cast(
            TransactionSettlementSynchronizedSata, self.synchronized_data
        )
        assert synchronized_data.missed_messages == expected_missed_messages


class TestFinalizationRound(BaseOnlyKeeperSendsRoundTest):
    """Test FinalizationRound."""

    _synchronized_data_class = TransactionSettlementSynchronizedSata
    _event_class = TransactionSettlementEvent
    _round_class = FinalizationRound

    @pytest.mark.parametrize(
        "tx_hashes_history, tx_digest, missed_messages, status, exit_event",
        (
            (
                "",
                "",
                {"test": 1},
                VerificationStatus.ERROR.value,
                TransactionSettlementEvent.CHECK_LATE_ARRIVING_MESSAGE,
            ),
            (
                "",
                "",
                {},
                VerificationStatus.ERROR.value,
                TransactionSettlementEvent.FINALIZATION_FAILED,
            ),
            (
                "t" * 66,
                "",
                {},
                VerificationStatus.VERIFIED.value,
                TransactionSettlementEvent.CHECK_HISTORY,
            ),
            (
                "t" * 66,
                "",
                {},
                VerificationStatus.ERROR.value,
                TransactionSettlementEvent.CHECK_HISTORY,
            ),
            (
                "",
                "",
                {},
                VerificationStatus.PENDING.value,
                TransactionSettlementEvent.FINALIZATION_FAILED,
            ),
            (
                "",
                "tx_digest" + "t" * 57,
                {},
                VerificationStatus.PENDING.value,
                TransactionSettlementEvent.DONE,
            ),
            (
                "t" * 66,
                "tx_digest" + "t" * 57,
                {},
                VerificationStatus.PENDING.value,
                TransactionSettlementEvent.DONE,
            ),
            (
                "t" * 66,
                "",
                {},
                VerificationStatus.INSUFFICIENT_FUNDS.value,
                TransactionSettlementEvent.INSUFFICIENT_FUNDS,
            ),
        ),
    )
    def test_finalization_round(
        self,
        tx_hashes_history: str,
        tx_digest: str,
        missed_messages: int,
        status: int,
        exit_event: TransactionSettlementEvent,
    ) -> None:
        """Runs tests."""
        keeper_retries = 2
        blacklisted_keepers = ""
        self.participants = frozenset([f"agent_{i}" + "-" * 35 for i in range(4)])
        keepers = deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35))
        self.synchronized_data = cast(
            TransactionSettlementSynchronizedSata,
            self.synchronized_data.update(
                participants=tuple(self.participants),
                missed_messages=missed_messages,
                tx_hashes_history=tx_hashes_history,
                keepers=get_keepers(keepers, keeper_retries),
                blacklisted_keepers=blacklisted_keepers,
            ),
        )

        sender = keepers[0]
        tx_hashes_history += (
            tx_digest
            if exit_event == TransactionSettlementEvent.DONE
            else tx_hashes_history
        )
        if status == VerificationStatus.INSUFFICIENT_FUNDS.value:
            popped = keepers.popleft()
            blacklisted_keepers += popped
            keeper_retries = 1

        test_round = self._round_class(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                keeper_payloads=FinalizationTxPayload(
                    sender=sender,
                    tx_data={
                        "status_value": status,
                        "serialized_keepers": get_keepers(keepers, keeper_retries),
                        "blacklisted_keepers": blacklisted_keepers,
                        "tx_hashes_history": tx_hashes_history,
                        "received_hash": bool(tx_digest),
                    },
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    tx_hashes_history=tx_hashes_history,
                    blacklisted_keepers=blacklisted_keepers,
                    keepers=get_keepers(keepers, keeper_retries),
                    keeper_retries=keeper_retries,
                    final_verification_status=VerificationStatus(status).value,
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.tx_hashes_history,
                    lambda _synchronized_data: _synchronized_data.blacklisted_keepers,
                    lambda _synchronized_data: _synchronized_data.keepers,
                    lambda _synchronized_data: _synchronized_data.keeper_retries,
                    lambda _synchronized_data: _synchronized_data.final_verification_status,
                ],
                exit_event=exit_event,
            )
        )

    def test_finalization_round_no_tx_data(self) -> None:
        """Test finalization round when `tx_data` is `None`."""
        keepers = deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35))
        keeper_retries = 2
        self.synchronized_data = cast(
            TransactionSettlementSynchronizedSata,
            self.synchronized_data.update(
                participants=tuple(f"agent_{i}" + "-" * 35 for i in range(4)),
                keepers=get_keepers(keepers, keeper_retries),
            ),
        )

        sender = keepers[0]

        test_round = self._round_class(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                keeper_payloads=FinalizationTxPayload(
                    sender=sender,
                    tx_data=None,
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data,
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.tx_hashes_history,
                    lambda _synchronized_data: _synchronized_data.blacklisted_keepers,
                    lambda _synchronized_data: _synchronized_data.keepers,
                    lambda _synchronized_data: _synchronized_data.keeper_retries,
                    lambda _synchronized_data: _synchronized_data.final_verification_status,
                ],
                exit_event=TransactionSettlementEvent.FINALIZATION_FAILED,
            )
        )


class TestCollectSignatureRound(BaseCollectDifferentUntilThresholdRoundTest):
    """Test CollectSignatureRound."""

    _synchronized_data_class = TransactionSettlementSynchronizedSata
    _event_class = TransactionSettlementEvent

    def test_run(
        self,
    ) -> None:
        """Runs tests."""

        test_round = CollectSignatureRound(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_signature(self.participants),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data,
                synchronized_data_attr_checks=[],
                exit_event=self._event_class.DONE,
            )
        )


class TestValidateTransactionRound(BaseValidateRoundTest):
    """Test ValidateRound."""

    test_class = ValidateTransactionRound
    _event_class = TransactionSettlementEvent
    _synchronized_data_class = TransactionSettlementSynchronizedSata


class TestCheckTransactionHistoryRound(BaseCollectSameUntilThresholdRoundTest):
    """Test CheckTransactionHistoryRound"""

    _event_class = TransactionSettlementEvent
    _synchronized_data_class = TransactionSettlementSynchronizedSata

    @pytest.mark.parametrize(
        "expected_status, expected_tx_hash, missed_messages, expected_event",
        (
            (
                "0000000000000000000000000000000000000000000000000000000000000001",
                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                {},
                TransactionSettlementEvent.DONE,
            ),
            (
                "0000000000000000000000000000000000000000000000000000000000000002",
                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                {},
                TransactionSettlementEvent.NEGATIVE,
            ),
            (
                "0000000000000000000000000000000000000000000000000000000000000003",
                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                {},
                TransactionSettlementEvent.NONE,
            ),
            (
                "0000000000000000000000000000000000000000000000000000000000000002",
                "b0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                {"test": 1},
                TransactionSettlementEvent.CHECK_LATE_ARRIVING_MESSAGE,
            ),
        ),
    )
    def test_run(
        self,
        expected_status: str,
        expected_tx_hash: str,
        missed_messages: int,
        expected_event: TransactionSettlementEvent,
    ) -> None:
        """Run tests."""
        keepers = get_keepers(deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)))
        self.synchronized_data.update(missed_messages=missed_messages, keepers=keepers)

        test_round = CheckTransactionHistoryRound(
            synchronized_data=self.synchronized_data,
        )

        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_check(
                    self.participants, expected_status, expected_tx_hash
                ),
                synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                    participant_to_check=CollectionRound.serialize_collection(
                        get_participant_to_check(
                            self.participants, expected_status, expected_tx_hash
                        )
                    ),
                    final_verification_status=int(expected_status),
                    tx_hashes_history=[expected_tx_hash],
                    keepers=keepers,
                    final_tx_hash="0xb0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9",
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.final_verification_status,
                    lambda _synchronized_data: _synchronized_data.final_tx_hash,
                    lambda _synchronized_data: _synchronized_data.keepers,
                ]
                if expected_event
                not in {
                    TransactionSettlementEvent.NEGATIVE,
                    TransactionSettlementEvent.CHECK_LATE_ARRIVING_MESSAGE,
                }
                else [
                    lambda _synchronized_data: _synchronized_data.final_verification_status,
                    lambda _synchronized_data: _synchronized_data.keepers,
                ],
                most_voted_payload=expected_status + expected_tx_hash,
                exit_event=expected_event,
            )
        )


class TestSynchronizeLateMessagesRound(BaseCollectNonEmptyUntilThresholdRound):
    """Test `SynchronizeLateMessagesRound`."""

    _event_class = TransactionSettlementEvent
    _synchronized_data_class = TransactionSettlementSynchronizedSata

    @pytest.mark.parametrize(
        "missed_messages, expected_event",
        (
            (
                {f"agent_{i}": 0 for i in range(4)},
                TransactionSettlementEvent.SUSPICIOUS_ACTIVITY,
            ),
            ({f"agent_{i}": 2 for i in range(4)}, TransactionSettlementEvent.DONE),
        ),
    )
    def test_runs(
        self, missed_messages: int, expected_event: TransactionSettlementEvent
    ) -> None:
        """Runs tests."""
        self.synchronized_data.update(missed_messages=missed_messages)
        test_round = SynchronizeLateMessagesRound(
            synchronized_data=self.synchronized_data,
        )
        late_arriving_tx_hashes = {
            p: "".join(("1" * TX_HASH_LENGTH, "2" * TX_HASH_LENGTH))
            for p in self.participants
        }
        test_round.required_block_confirmations = 0
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_late_arriving_tx_hashes(
                    self.participants
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.update(
                    late_arriving_tx_hashes=late_arriving_tx_hashes,
                    suspects=tuple()
                    if expected_event == TransactionSettlementEvent.DONE
                    else tuple(sorted(late_arriving_tx_hashes.keys())),
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.late_arriving_tx_hashes,
                    lambda _synchronized_data: _synchronized_data.suspects,
                ],
                exit_event=expected_event,
            )
        )

    @pytest.mark.parametrize("correct_serialization", (True, False))
    def test_check_payload(self, correct_serialization: bool) -> None:
        """Test the `check_payload` method."""

        test_round = SynchronizeLateMessagesRound(
            synchronized_data=self.synchronized_data,
        )
        sender = list(test_round.accepting_payloads_from).pop()
        hash_length = TX_HASH_LENGTH
        if not correct_serialization:
            hash_length -= 1
        tx_hashes = "0" * hash_length
        payload = SynchronizeLateMessagesPayload(sender=sender, tx_hashes=tx_hashes)

        if correct_serialization:
            test_round.check_payload(payload)
            return

        with pytest.raises(
            TransactionNotValidError, match="Expecting serialized data of chunk size"
        ):
            test_round.check_payload(payload)

        with pytest.raises(
            ABCIAppInternalError, match="Expecting serialized data of chunk size"
        ):
            test_round.process_payload(payload)
        assert payload not in test_round.collection


def test_synchronized_datas() -> None:
    """Test SynchronizedData."""

    participants = get_participants()
    participant_to_randomness = get_participant_to_randomness(participants, 1)
    participant_to_randomness_serialized = CollectionRound.serialize_collection(
        participant_to_randomness
    )
    most_voted_randomness = get_most_voted_randomness()
    participant_to_selection = get_participant_to_selection(participants, "test")
    participant_to_selection_serialized = CollectionRound.serialize_collection(
        participant_to_selection
    )
    safe_contract_address = get_safe_contract_address()
    most_voted_tx_hash = get_most_voted_tx_hash()
    participant_to_signature = get_participant_to_signature(participants)
    participant_to_signature_serialized = CollectionRound.serialize_collection(
        participant_to_signature
    )
    final_tx_hash = get_final_tx_hash()
    actual_keeper_randomness = int(most_voted_randomness, base=16) / MAX_INT_256
    late_arriving_tx_hashes_serialized = get_late_arriving_tx_hashes_serialized()
    late_arriving_tx_hashes_deserialized = get_late_arriving_tx_hashes_deserialized()
    keepers = get_keepers(deque(("agent_1" + "-" * 35, "agent_3" + "-" * 35)))
    expected_keepers = deque(["agent_1" + "-" * 35, "agent_3" + "-" * 35])

    # test `keeper_retries` property when no `keepers` are set.
    synchronized_data_____ = TransactionSettlementSynchronizedSata(
        AbciAppDB(setup_data=dict())
    )
    assert synchronized_data_____.keepers == deque()
    assert synchronized_data_____.keeper_retries == 0

    synchronized_data_____ = TransactionSettlementSynchronizedSata(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    all_participants=tuple(participants),
                    participants=tuple(participants),
                    consensus_threshold=3,
                    participant_to_randomness=participant_to_randomness_serialized,
                    most_voted_randomness=most_voted_randomness,
                    participant_to_selection=participant_to_selection_serialized,
                    safe_contract_address=safe_contract_address,
                    most_voted_tx_hash=most_voted_tx_hash,
                    participant_to_signature=participant_to_signature_serialized,
                    final_tx_hash=final_tx_hash,
                    late_arriving_tx_hashes=late_arriving_tx_hashes_serialized,
                    keepers=keepers,
                    blacklisted_keepers="t" * 42,
                )
            ),
        )
    )
    assert (
        abs(synchronized_data_____.keeper_randomness - actual_keeper_randomness) < 1e-10
    )  # avoid equality comparisons between floats
    assert synchronized_data_____.most_voted_randomness == most_voted_randomness
    assert synchronized_data_____.safe_contract_address == safe_contract_address
    assert synchronized_data_____.most_voted_tx_hash == most_voted_tx_hash
    assert synchronized_data_____.participant_to_randomness == participant_to_randomness
    assert synchronized_data_____.participant_to_selection == participant_to_selection
    assert synchronized_data_____.participant_to_signature == participant_to_signature
    assert synchronized_data_____.final_tx_hash == final_tx_hash
    assert (
        synchronized_data_____.late_arriving_tx_hashes
        == late_arriving_tx_hashes_deserialized
    )
    assert synchronized_data_____.keepers == expected_keepers
    assert synchronized_data_____.keeper_retries == 1
    assert (
        synchronized_data_____.most_voted_keeper_address == expected_keepers.popleft()
    )
    assert synchronized_data_____.keepers_threshold_exceeded
    assert synchronized_data_____.blacklisted_keepers == {"t" * 42}
    updated_synchronized_data = synchronized_data_____.create()
    assert updated_synchronized_data.blacklisted_keepers == set()


class TestResetRound(BaseCollectSameUntilThresholdRoundTest):
    """Test ResetRound."""

    _synchronized_data_class = TransactionSettlementSynchronizedSata
    _event_class = TransactionSettlementEvent

    def test_runs(
        self,
    ) -> None:
        """Runs tests."""
        randomness = DUMMY_RANDOMNESS
        synchronized_data = self.synchronized_data.update(
            most_voted_randomness=randomness,
            late_arriving_tx_hashes={},
            keepers="",
        )
        synchronized_data._db._cross_period_persisted_keys = frozenset(
            {"most_voted_randomness"}
        )
        test_round = ResetRound(synchronized_data=synchronized_data)
        next_period_count = 1
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=get_participant_to_period_count(
                    self.participants, next_period_count
                ),
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data.create(),
                synchronized_data_attr_checks=[],  # [lambda _synchronized_data: _synchronized_data.participants],
                most_voted_payload=next_period_count,
                exit_event=self._event_class.DONE,
            )
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=unused-import

"""Test the models.py module of the skill."""
from typing import Dict
from unittest.mock import MagicMock

import pytest
import yaml
from aea.exceptions import AEAEnforceError

from packages.valory.skills.abstract_round_abci.test_tools.base import DummyContext
from packages.valory.skills.transaction_settlement_abci.models import (
    TransactionParams,
    _MINIMUM_VALIDATE_TIMEOUT,
)
from packages.valory.skills.transaction_settlement_abci.tests import PACKAGE_DIR


class TestTransactionParams:  # pylint: disable=too-few-public-methods
    """Test TransactionParams class."""

    default_config: Dict

    def setup_class(self) -> None:
        """Read the default config only once."""
        skill_yaml = PACKAGE_DIR / "skill.yaml"
        with open(skill_yaml, "r", encoding="utf-8") as skill_file:
            skill = yaml.safe_load(skill_file)
        self.default_config = skill["models"]["params"]["args"]

    def test_ensure_validate_timeout(  # pylint: disable=no-self-use
        self,
    ) -> None:
        """Test that `_ensure_validate_timeout` raises when `validate_timeout` is lower than the allowed minimum."""
        dummy_value = 0
        mock_args, mock_kwargs = (
            MagicMock(),
            {
                **self.default_config,
                "validate_timeout": dummy_value,
                "skill_context": DummyContext(),
            },
        )
        with pytest.raises(
            expected_exception=AEAEnforceError,
            match=f"`validate_timeout` must be greater than or equal to {_MINIMUM_VALIDATE_TIMEOUT}",
        ):
            TransactionParams(mock_args, **mock_kwargs)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'test_ipfs_behaviour' skill."""
import sys
from abc import ABC
from typing import Generator, Set, Type, cast

from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.io_.store import SupportedFiletype
from packages.valory.skills.test_ipfs_abci.rounds import IpfsRound, IpfsTestAbciApp


class DummyIpfsBehaviour(BaseBehaviour, ABC):
    """A dummy behaviour that stores and retrieves data from IPFS."""

    matching_round = IpfsRound

    def async_act(self) -> Generator:
        """Do the action."""
        dummy_object = {
            "dummy_k1": "dummy_v1",
            "dummy_k2": "dummy_v2",
            "dummy_k3": "dummy_v3",
        }

        filename = "testfile.json"
        ipfs_hash = yield from self.send_to_ipfs(
            filename,
            {filename: dummy_object},
            filetype=SupportedFiletype.JSON,
            multiple=True,
        )
        ipfs_hash = cast(str, ipfs_hash)
        received_object = yield from self.get_from_ipfs(
            ipfs_hash=ipfs_hash, filetype=SupportedFiletype.JSON
        )
        if received_object == dummy_object:
            self.context.logger.info("Single object uploading & downloading works.")

        multiple_objects = {
            "test1.json": dummy_object,
            "test2.json": dummy_object,
            "test3.json": dummy_object,
        }
        dir_name = "test_dir"
        ipfs_hash = yield from self.send_to_ipfs(
            dir_name,
            multiple_objects,
            filetype=SupportedFiletype.JSON,
            multiple=True,
        )
        ipfs_hash = cast(str, ipfs_hash)
        received_object = yield from self.get_from_ipfs(
            ipfs_hash=ipfs_hash, filetype=SupportedFiletype.JSON
        )
        if received_object == multiple_objects:
            self.context.logger.info("Multiple object uploading & downloading works.")
        else:
            self.context.logger.info(received_object)
            self.context.logger.info(multiple_objects)
        # give some time to
        yield from self.sleep(10)
        sys.exit()


class TestAbciConsensusBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the ipfs test abci app."""

    initial_behaviour_cls = DummyIpfsBehaviour
    abci_app_cls = IpfsTestAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        DummyIpfsBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Watcher script and wrapper container for agent."""

import os
import shutil
import signal
import subprocess  # nosec
import sys
import time
from pathlib import Path
from typing import Optional

import requests
from watchdog.events import EVENT_TYPE_CLOSED, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer


ID = os.environ.get("ID")
ROOT = "/home/ubuntu"
AGENT_DIR = ROOT + "/agent"
PACKAGES_PATH = "/home/ubuntu/packages"
OPEN_AEA_PATH = "/open-aea"
BASE_START_FILE = "/home/ubuntu/start.sh"
TENDERMINT_COM_URL = os.environ.get("TENDERMINT_COM_URL", f"http://node{ID}:8080")


def write(line: str) -> None:
    """Write to console."""
    sys.stdout.write(line)
    sys.stdout.write("\n")
    sys.stdout.flush()


def call_vote() -> None:
    """
    Call vote.

    Since there's a lot of resource sharing between docker containers one of the
    environments can fallback during `base_setup` so to make sure there's no error
    caused by one of the agents left behind this method will help.
    """
    write("Calling vote.")
    with open(f"/logs/{ID}.vote", "w+") as fp:
        fp.write(str(ID))


class AEARunner:
    """AEA Runner."""

    process: Optional[subprocess.Popen]  # nosec

    def __init__(self) -> None:
        """Initialize runner."""

        self.process = None

    @staticmethod
    def restart_tendermint() -> None:
        """Restart respective tendermint node."""
        write("Restarting Tendermint.")
        try:
            response = requests.get(TENDERMINT_COM_URL + "/hard_reset")
            if response.status_code != 200:
                write("Tendermint node not yet available.")
        except requests.exceptions.ConnectionError:
            write("Tendermint node not yet available.")

    def start(
        self,
    ) -> None:
        """Start AEA process."""

        if self.process is not None:
            return
        write("Starting Agent.")
        os.chdir(ROOT)
        if Path(AGENT_DIR).exists():
            shutil.rmtree(AGENT_DIR)
        self.process = subprocess.Popen(  # nosec
            ["/bin/bash", BASE_START_FILE], preexec_fn=os.setsid
        )

    def stop(
        self,
    ) -> None:
        """Stop AEA process."""
        if self.process is None:
            return
        write("Stopping Agent.")
        os.killpg(
            os.getpgid(self.process.pid), signal.SIGTERM
        )  # kills process instantly compared to process.terminate
        self.process = None


class EventHandler(FileSystemEventHandler):
    """Handle file updates."""

    def __init__(
        self, aea_runner: AEARunner, fingerprint_on_restart: bool = True
    ) -> None:
        """Initialize object."""
        super().__init__()
        self.aea = aea_runner
        self.fingerprint_on_restart = fingerprint_on_restart

    @staticmethod
    def fingerprint_item(src_path: str) -> None:
        """Fingerprint items."""
        cwd = os.curdir
        *_path, vendor, item_type, item_name, _ = src_path.split(os.path.sep)
        vendor_dir_str = os.path.sep.join([*_path, vendor])
        os.chdir(vendor_dir_str)
        subprocess.call(  # nosec
            [
                "python3",
                "-m",
                "aea.cli",
                "fingerprint",
                item_type[:-1],
                f"{vendor}/{item_name}",
            ]
        )
        os.chdir(cwd)

    @staticmethod
    def clean_up() -> None:
        """Clean up from previous run."""
        os.chdir(ROOT)
        if Path(AGENT_DIR).exists():
            write("removing aea dir.")
            shutil.rmtree(AGENT_DIR)

    def on_any_event(self, event: FileSystemEvent) -> None:
        """This method reloads the agent when a change is detected in *.py file."""
        if (
            not event.is_directory
            and event.event_type == EVENT_TYPE_CLOSED
            and event.src_path.endswith(".py")
        ):
            write("Change detected.")
            self.clean_up()
            if self.fingerprint_on_restart:
                self.fingerprint_item(event.src_path)

            self.aea.stop()
            self.aea.restart_tendermint()
            self.aea.start()


if __name__ == "__main__":
    aea_runner = AEARunner()

    package_observer = Observer()
    package_observer.schedule(
        EventHandler(aea_runner=aea_runner), PACKAGES_PATH, recursive=True
    )

    open_aea_observer = Observer()
    open_aea_observer.schedule(
        EventHandler(aea_runner=aea_runner, fingerprint_on_restart=False),
        OPEN_AEA_PATH,
        recursive=True,
    )

    try:
        aea_runner.start()
        package_observer.start()
        open_aea_observer.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        aea_runner.stop()
        package_observer.stop()
        open_aea_observer.stop()

    open_aea_observer.join()
    package_observer.join()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for abstract_round_abci/test_tools/common.py"""

from typing import Type, Union, cast

import pytest

from packages.valory.skills.abstract_round_abci.test_tools.common import (
    BaseRandomnessBehaviourTest,
    BaseSelectKeeperBehaviourTest,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci import (
    PATH_TO_SKILL,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.behaviours import (
    DummyFinalBehaviour,
    DummyKeeperSelectionBehaviour,
    DummyRandomnessBehaviour,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.rounds import (
    Event,
)
from packages.valory.skills.abstract_round_abci.tests.test_tools.base import (
    FSMBehaviourTestToolSetup,
)


class BaseCommonBaseCaseTestSetup(FSMBehaviourTestToolSetup):
    """BaseRandomnessBehaviourTestSetup"""

    test_cls: Type[Union[BaseRandomnessBehaviourTest, BaseSelectKeeperBehaviourTest]]

    def set_done_event(self) -> None:
        """Set done_event"""
        self.test_cls.done_event = Event.DONE

    def set_next_behaviour_class(self, next_behaviour_class: Type) -> None:
        """Set next_behaviour_class"""
        self.test_cls.next_behaviour_class = next_behaviour_class


class TestBaseRandomnessBehaviourTestSetup(BaseCommonBaseCaseTestSetup):
    """Test BaseRandomnessBehaviourTest setup."""

    test_cls: Type[BaseRandomnessBehaviourTest] = BaseRandomnessBehaviourTest

    def set_randomness_behaviour_class(self) -> None:
        """Set randomness_behaviour_class"""
        self.test_cls.randomness_behaviour_class = DummyRandomnessBehaviour  # type: ignore

    def test_setup_randomness_behaviour_class_not_set(self) -> None:
        """Test setup randomness_behaviour_class not set."""

        self.set_path_to_skill()
        test_instance = cast(BaseRandomnessBehaviourTest, self.setup_test_cls())
        expected = f"'{self.test_cls.__name__}' object has no attribute 'randomness_behaviour_class'"
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_randomness_behaviour()

    def test_setup_done_event_not_set(self) -> None:
        """Test setup done_event = Event.DONE not set."""

        self.set_path_to_skill()
        self.set_randomness_behaviour_class()

        test_instance = cast(BaseRandomnessBehaviourTest, self.setup_test_cls())
        expected = f"'{self.test_cls.__name__}' object has no attribute 'done_event'"
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_randomness_behaviour()

    def test_setup_next_behaviour_class_not_set(self) -> None:
        """Test setup next_behaviour_class not set."""

        self.set_path_to_skill()
        self.set_randomness_behaviour_class()
        self.set_done_event()

        test_instance = cast(BaseRandomnessBehaviourTest, self.setup_test_cls())
        expected = (
            f"'{self.test_cls.__name__}' object has no attribute 'next_behaviour_class'"
        )
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_randomness_behaviour()

    def test_successful_setup_randomness_behaviour_test(self) -> None:
        """Test successful setup of the test class inheriting from BaseRandomnessBehaviourTest."""

        self.set_path_to_skill()
        self.set_randomness_behaviour_class()
        self.set_done_event()
        self.set_next_behaviour_class(DummyKeeperSelectionBehaviour)
        test_instance = cast(BaseRandomnessBehaviourTest, self.setup_test_cls())
        test_instance.test_randomness_behaviour()


class TestBaseRandomnessBehaviourTestRunning(BaseRandomnessBehaviourTest):
    """Test TestBaseRandomnessBehaviourTestRunning running."""

    path_to_skill = PATH_TO_SKILL
    randomness_behaviour_class = DummyRandomnessBehaviour
    next_behaviour_class = DummyKeeperSelectionBehaviour
    done_event = Event.DONE


class TestBaseSelectKeeperBehaviourTestSetup(BaseCommonBaseCaseTestSetup):
    """Test BaseRandomnessBehaviourTest setup."""

    test_cls: Type[BaseSelectKeeperBehaviourTest] = BaseSelectKeeperBehaviourTest

    def set_select_keeper_behaviour_class(self) -> None:
        """Set select_keeper_behaviour_class"""
        self.test_cls.select_keeper_behaviour_class = DummyKeeperSelectionBehaviour  # type: ignore

    def test_setup_select_keeper_behaviour_class_not_set(self) -> None:
        """Test setup select_keeper_behaviour_class not set."""

        self.set_path_to_skill()
        test_instance = cast(BaseSelectKeeperBehaviourTest, self.setup_test_cls())
        expected = f"'{self.test_cls.__name__}' object has no attribute 'select_keeper_behaviour_class'"
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_select_keeper_preexisting_keeper()

    def test_setup_done_event_not_set(self) -> None:
        """Test setup done_event = Event.DONE not set."""

        self.set_path_to_skill()
        self.set_select_keeper_behaviour_class()

        test_instance = cast(BaseSelectKeeperBehaviourTest, self.setup_test_cls())
        expected = f"'{self.test_cls.__name__}' object has no attribute 'done_event'"
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_select_keeper_preexisting_keeper()

    def test_setup_next_behaviour_class_not_set(self) -> None:
        """Test setup next_behaviour_class not set."""

        self.set_path_to_skill()
        self.set_select_keeper_behaviour_class()
        self.set_done_event()

        test_instance = cast(BaseSelectKeeperBehaviourTest, self.setup_test_cls())
        expected = (
            f"'{self.test_cls.__name__}' object has no attribute 'next_behaviour_class'"
        )
        with pytest.raises(AttributeError, match=expected):
            test_instance.test_select_keeper_preexisting_keeper()

    def test_successful_setup_select_keeper_behaviour_test(self) -> None:
        """Test successful setup of the test class inheriting from BaseSelectKeeperBehaviourTest."""

        self.set_path_to_skill()
        self.set_select_keeper_behaviour_class()
        self.set_done_event()
        self.set_next_behaviour_class(DummyFinalBehaviour)
        test_instance = cast(BaseSelectKeeperBehaviourTest, self.setup_test_cls())
        test_instance.test_select_keeper_preexisting_keeper()


class TestBaseSelectKeeperBehaviourTestRunning(BaseSelectKeeperBehaviourTest):
    """Test BaseSelectKeeperBehaviourTest running."""

    path_to_skill = PATH_TO_SKILL
    select_keeper_behaviour_class = DummyKeeperSelectionBehaviour
    next_behaviour_class = DummyFinalBehaviour
    done_event = Event.DONE



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for packages command"""

from collections import namedtuple
from typing import Any
from unittest import mock

from autonomy.cli.packages import get_package_manager

from tests.test_autonomy.test_cli.base import BaseCliTest


class TestPackages(BaseCliTest):
    """Test packages command"""

    cli_options = ("packages", "lock")

    def test_lock(self) -> None:
        """Test lock"""

        target = "autonomy.cli.packages.get_package_manager"
        with mock.patch(target, return_value=mock.MagicMock()) as m:
            with mock.patch("sys.exit") as sys_exit_mock:
                result = self.run_cli()
                assert result.exit_code == 0
                assert "Updating hashes..." in result.stdout
                m.assert_called_once()
                assert "Done" in result.stdout
                sys_exit_mock.assert_called_with(0)

    def test_lock_check(self) -> None:
        """Test lock --check"""

        ctr = 0

        def side_effect(*args: Any, **kwargs: Any) -> Any:
            nonlocal ctr
            if ctr == 0:
                ctr += 1
                return get_package_manager(*args, **kwargs)
            return mock.MagicMock()

        target = "autonomy.cli.packages.get_package_manager"
        with mock.patch(target, side_effect=side_effect):
            with mock.patch("sys.exit") as sys_exit_mock:
                result = self.run_cli(("--check",))
                assert result.exit_code == 0
                assert "Verifying packages.json" in result.stdout
                assert "Verification successful" in result.stdout
                sys_exit_mock.assert_called_with(0)

    def test_lock_check_fail(self) -> None:
        """Test lock --check failure"""

        DummyPackageManager = namedtuple("DummyPackageManager", "verify")
        return_value = DummyPackageManager(lambda *_, **__: 1)
        target = "autonomy.cli.packages.get_package_manager"
        with mock.patch(target, return_value=return_value):
            result = self.run_cli(("--check",))
            assert result.exit_code == 1
            assert "Verification failed." in result.stdout



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test hash command group."""


import json
import shutil
from pathlib import Path
from typing import Dict, Tuple

import _strptime  # noqa  # pylint: disable=unsed-import

from autonomy.cli import cli
from autonomy.configurations.loader import load_service_config

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestHashAll(BaseCliTest):
    """Test `hash-all` command."""

    packages_dir: Path
    cli_options: Tuple[str, ...] = (
        "hash",
        "all",
    )

    def setup(self) -> None:
        """Setup class."""

        super().setup()

        self.packages_dir = self.t / "packages"
        shutil.copytree(ROOT_DIR / "packages", self.packages_dir)

    def load_hashes(
        self,
    ) -> Dict:
        """Load hashes from CSV file."""

        hashes_file = self.packages_dir / "packages.json"
        with open(str(hashes_file), "r") as file:
            hashes = json.load(file)

        return hashes

    def test_service_hashing(
        self,
    ) -> None:
        """Check if `hash-all` updates agent hashes in service configs properly."""

        service_name = "counter"
        result = self.run_cli(("--packages-dir", str(self.packages_dir)))

        assert result.exit_code == 0

        service_path = self.packages_dir / "valory" / "services" / service_name
        service_config = load_service_config(service_path)
        hashes = self.load_hashes()
        key = f"agent/valory/{service_name}/0.1.0"

        assert key in hashes["dev"], (
            hashes,
            service_config.agent,
        )
        assert hashes["dev"][key] == service_config.agent.hash

        result = self.cli_runner.invoke(
            cli=cli, args=(*("packages", "lock"), *("--check",))
        )

        assert result.exit_code == 0, result.output
        assert "Verifying packages.json" in result.output, result.output


class TestHashOne(BaseCliTest):
    """Test `hash one` command."""

    file: Path
    original_hash: str = "bafybeihptu54tiz3tilyueo2wv5qpwts6lcswwhlu4uh2kculu7q4f4kuq"
    cli_options: Tuple[str, ...] = (
        "hash",
        "one",
    )

    def setup(self) -> None:
        """Setup class."""

        super().setup()

        self.file = self.t / "some_file.txt"
        self.file.write_text("Hello, World!")

    def test_one(
        self,
    ) -> None:
        """Test `hash one` command."""

        result = self.run_cli((str(self.file),))

        assert result.exit_code == 0, result.output
        assert self.original_hash in result.output, result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpers for minting components"""

import datetime
import time
from math import ceil
from typing import Callable, Dict, List, Optional, Tuple

from aea.crypto.base import Crypto, LedgerApi
from requests.exceptions import ConnectionError as RequestsConnectionError

from autonomy.chain.base import UnitType, registry_contracts
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.constants import (
    AGENT_REGISTRY_CONTRACT,
    COMPONENT_REGISTRY_CONTRACT,
    EVENT_VERIFICATION_TIMEOUT,
    REGISTRIES_MANAGER_CONTRACT,
    SERVICE_MANAGER_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
)
from autonomy.chain.exceptions import (
    ComponentMintFailed,
    FailedToRetrieveTokenId,
    InvalidMintParameter,
)


DEFAULT_NFT_IMAGE_HASH = "bafybeiggnad44tftcrenycru2qtyqnripfzitv5yume4szbkl33vfd4abm"


def transact(ledger_api: LedgerApi, crypto: Crypto, tx: Dict) -> Dict:
    """Make a transaction and return a receipt"""

    tx_signed = crypto.sign_transaction(transaction=tx)
    tx_digest = ledger_api.send_signed_transaction(tx_signed=tx_signed)

    return ledger_api.get_transaction_receipt(tx_digest=tx_digest)


def sort_service_dependency_metadata(
    agent_ids: List[int],
    number_of_slots_per_agents: List[int],
    cost_of_bond_per_agent: List[int],
) -> Tuple[List[int], ...]:
    """Sort service dependencies and their respective parameters"""

    ids_sorted = []
    slots_sorted = []
    securities_sorted = []

    for idx, n_slots, b_cost in sorted(
        zip(agent_ids, number_of_slots_per_agents, cost_of_bond_per_agent),
        key=lambda x: x[0],
    ):
        ids_sorted.append(idx)
        slots_sorted.append(n_slots)
        securities_sorted.append(b_cost)

    return ids_sorted, slots_sorted, securities_sorted


def wait_for_component_to_mint(
    token_retriever: Callable[[], Optional[int]],
    timeout: Optional[float] = None,
    sleep: float = 1.0,
) -> int:
    """Wait for service activation."""

    timeout = timeout or EVENT_VERIFICATION_TIMEOUT
    deadline = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
    while datetime.datetime.now() < deadline:
        token_id = token_retriever()
        if token_id is not None:
            return token_id
        time.sleep(sleep)

    raise TimeoutError("Could not retrieve the token in given time limit.")


def mint_component(  # pylint: disable=too-many-arguments
    ledger_api: LedgerApi,
    crypto: Crypto,
    metadata_hash: str,
    component_type: UnitType,
    chain_type: ChainType,
    owner: Optional[str] = None,
    dependencies: Optional[List[int]] = None,
    timeout: Optional[float] = None,
) -> Optional[int]:
    """Publish component on-chain."""

    if dependencies is not None:
        dependencies = sorted(set(dependencies))

    try:
        owner = ledger_api.api.toChecksumAddress(owner or crypto.address)
    except ValueError as e:  # pragma: nocover
        raise ComponentMintFailed(f"Invalid owner address {owner}") from e

    try:
        tx = registry_contracts.registries_manager.get_create_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                REGISTRIES_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=owner,
            sender=crypto.address,
            component_type=component_type,
            metadata_hash=metadata_hash,
            dependencies=dependencies,
            raise_on_try=True,
        )
        transact(
            ledger_api=ledger_api,
            crypto=crypto,
            tx=tx,
        )
    except RequestsConnectionError as e:
        raise ComponentMintFailed("Cannot connect to the given RPC") from e

    try:
        if component_type == UnitType.COMPONENT:

            def token_retriever() -> bool:
                """Retrieve token"""
                return registry_contracts.component_registry.filter_token_id_from_emitted_events(
                    ledger_api=ledger_api,
                    contract_address=ContractConfigs.get(
                        COMPONENT_REGISTRY_CONTRACT.name
                    ).contracts[chain_type],
                    metadata_hash=metadata_hash,
                )

        else:

            def token_retriever() -> bool:
                """Retrieve token"""
                return registry_contracts.agent_registry.filter_token_id_from_emitted_events(
                    ledger_api=ledger_api,
                    contract_address=ContractConfigs.get(
                        AGENT_REGISTRY_CONTRACT.name
                    ).contracts[chain_type],
                    metadata_hash=metadata_hash,
                )

        return wait_for_component_to_mint(
            token_retriever=token_retriever, timeout=timeout
        )

    except RequestsConnectionError as e:
        raise FailedToRetrieveTokenId(
            "Connection interrupted while waiting for the unitId emit event"
        ) from e
    except TimeoutError as e:
        raise FailedToRetrieveTokenId(str(e)) from e


def mint_service(  # pylint: disable=too-many-arguments
    ledger_api: LedgerApi,
    crypto: Crypto,
    metadata_hash: str,
    chain_type: ChainType,
    agent_ids: List[int],
    number_of_slots_per_agent: List[int],
    cost_of_bond_per_agent: List[int],
    threshold: int,
    owner: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Optional[int]:
    """Publish component on-chain."""

    if len(agent_ids) == 0:
        raise InvalidMintParameter("Please provide at least one agent id")

    if len(number_of_slots_per_agent) == 0:
        raise InvalidMintParameter("Please for provide number of slots for agents")

    if len(cost_of_bond_per_agent) == 0:
        raise InvalidMintParameter("Please for provide cost of bond for agents")

    if (
        len(agent_ids) != len(number_of_slots_per_agent)
        or len(agent_ids) != len(cost_of_bond_per_agent)
        or len(number_of_slots_per_agent) != len(cost_of_bond_per_agent)
    ):
        raise InvalidMintParameter(
            "Make sure the number of agent ids, number of slots for agents and cost of bond for agents match"
        )

    if any(map(lambda x: x == 0, number_of_slots_per_agent)):
        raise InvalidMintParameter("Number of slots cannot be zero")

    if any(map(lambda x: x == 0, cost_of_bond_per_agent)):
        raise InvalidMintParameter("Cost of bond cannot be zero")

    number_of_agent_instances = sum(number_of_slots_per_agent)
    if threshold < (ceil((number_of_agent_instances * 2 + 1) / 3)):
        raise InvalidMintParameter(
            "The threshold value should at least be greater than or equal to ceil((n * 2 + 1) / 3), "
            "n is total number of agent instances in the service"
        )

    (
        agent_ids,
        number_of_slots_per_agent,
        cost_of_bond_per_agent,
    ) = sort_service_dependency_metadata(
        agent_ids=agent_ids,
        number_of_slots_per_agents=number_of_slots_per_agent,
        cost_of_bond_per_agent=cost_of_bond_per_agent,
    )

    agent_params = [
        [n, c] for n, c in zip(number_of_slots_per_agent, cost_of_bond_per_agent)
    ]

    try:
        owner = ledger_api.api.toChecksumAddress(owner or crypto.address)
    except ValueError as e:
        raise ComponentMintFailed(f"Invalid owner address {owner}") from e

    try:
        tx = registry_contracts.service_manager.get_create_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=owner,
            sender=crypto.address,
            metadata_hash=metadata_hash,
            agent_ids=agent_ids,
            agent_params=agent_params,
            threshold=threshold,
            raise_on_try=True,
        )
        transact(
            ledger_api=ledger_api,
            crypto=crypto,
            tx=tx,
        )
    except RequestsConnectionError as e:
        raise ComponentMintFailed("Cannot connect to the given RPC") from e

    try:

        def token_retriever() -> bool:
            """Retrieve token"""
            return (
                registry_contracts.service_registry.filter_token_id_from_emitted_events(
                    ledger_api=ledger_api,
                    contract_address=ContractConfigs.get(
                        SERVICE_REGISTRY_CONTRACT.name
                    ).contracts[chain_type],
                )
            )

        return wait_for_component_to_mint(
            token_retriever=token_retriever, timeout=timeout
        )

    except RequestsConnectionError as e:
        raise FailedToRetrieveTokenId(
            "Connection interrupted while waiting for the unitId emit event"
        ) from e
    except TimeoutError as e:
        raise FailedToRetrieveTokenId(str(e)) from e
